/*
 * $Id: charsets.c,v 1.1 1995/04/01 07:50:52 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Necessary Character set conversions.
 * Included are:
 *	o Petscii <=> Ascii
 *	o Petscii  => ISO Latin/1
 *
 *
 * Written by
 *   Jouko Valta  (jopi@stekt.oulu.fi)
 *
 *
 * $Log: charsets.c,v $
 * Revision 1.1  1995/04/01  07:50:52  jopi
 * Initial revision
 *
 *
 *
 */


#include <stdio.h>
#include <ctype.h>

#include "charsets.h"

typedef unsigned char BYTE;


#ifndef __CEXTRACT__

extern void petconvstring ( char *c, int dir );
extern int  p_toascii ( int c );
extern int  p_topetcii ( int c );

extern int  set_locale ( char *localename );
extern int  p_toichar ( int lang, int lwcase, BYTE c );
extern int  str_to_lang ( char *name );

#endif /* __CEXTRACT__ */


void    petconvstring(char *c, int dir)
{
    char *p = c;

    if (dir)
	for (; *p; p++)
	    *p = p_topetcii(*p);
    else
	for (; *p; p++)
	    *p = p_toascii(*p);
}


int   p_toascii(int c)
{
    if ((c >= 'A') && (c <= 0x7e))
	return (c ^ 0x20);
    else if (isprint(c)) {
	return (c);
    }
    else {
	if (c >= 0xc0 && c < 0xe0)
	    return (c - 0x80);
	else if (c == 0xff)
	    return (0x7e);
	else if (c == 0x0d)
	    return ('\n');
	else if (c == 0xa0)	/* Shifted Space */
	    return (' ');
    }
    /*return (c);*/
    return ((isprint(c) ? c : '.'));
}


int   p_topetcii(int c)
{
    if (c == 0x7e)		/*  '~' is ASCII for 'pi' */
	return (0xff);

    else if ((c >= 0x5b) && (c <= 0x7e))
	return (c ^ 0x20);

    else if ((c >= 'A') && (c <= 'Z'))
	return (c | 0x80);

    return (c);
}


/* ------------------------------------------------------------------------- */

/*
 * Commodore's International character sets
 */

#define NUM_LANGUAGES	10	/* 0 - 9 */

static char *langnames[] = {	/* International Charsets */
	"commodore",	/* CBM  */
	"usa"		/* US   */
	"germany",	/* GER  */
	"denmark-1",	/* DEN1 */
	"france",	/* FRA  */
	"sweden-1",	/* SWE1 */
	"italy",	/* ITA  */
	"spain",	/* SPA  */
	"denmark-2",	/* DEN2 */
	"sweden-2",	/* SWE2 */
			/* UK   */
	""
};


/* 7-bit */

/* ASCII/ISO 646 family */
static BYTE ASCextens1[][6] = {		/* graphics/uppercase */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* CBM  */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* US   */
    { 0x5b,0x5c,0x55,0x5e,0x5f,0xc4 },	/* GER  */
    { 0x7b,0x7c,0x7d,0x5e,0x5f,0xc4 },	/* DEN1 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* FRA  */
    { 0x7b,0x7c,0x7d,0x5e,0x5f,0xc4 },	/* SWE1 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* ITA  */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* SPA  */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* DEN2 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* SWE2 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 }	/* UK   */
};

static BYTE ASCextens2[][5] = {		/* uppercase/lowercase */
    { 0x2b,0x7c,0x7c,0x26,0x5c },	/* CBM  */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* US   */
    { 0x7b,0x7c,0x75,0x7e,0x23 },	/* GER  */
    { 0x5b,0x5c,0x5d,0x7e,0x7f },	/* DEN1 */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* FRA  */
    { 0x5b,0x5c,0x5d,0x7e,0x7f },	/* SWE1 */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* ITA  */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* SPA  */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* DEN2 */
    { 0x7b,0x7c,0x7d,0x7e,0x23 },	/* SWE2 */
    { 0x7b,0x7c,0x7d,0x7e,0x7f }	/* UK   */
};


/* 8-bit */

/* ANSI/ISO 8859-1 */
static BYTE ISOextens1[][6] = {		/* graphics/uppercase */
    { 0x5b,0xa3,0x5d,0x5e,0x5f,0xc4 },	/* CBM  */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* US   */
    { 0xc4,0xd6,0xdc,0x5e,0x5f,0xc4 },	/* GER  */
    { 0xe6,0xf8,0xe5,0x5e,0x5f,0xc4 },	/* DEN1 */
    { 0xb0,0xe7,0xa7,0x5e,0x5f,0xc4 },	/* FRA  */
    { 0xe4,0xf6,0xe5,0x5e,0x5f,0xc4 },	/* SWE1 */
    { 0xb0,0x5c,0xe9,0x5e,0x5f,0xc4 },	/* ITA  */
    { 0xa1,0xd1,0xbf,0x5e,0x5f,0xc4 },	/* SPA  */
    { 0xc6,0xd8,0xc5,0x5e,0x5f,0xc4 },	/* DEN2 */
    { 0xc4,0xd6,0xc5,0x5e,0x5f,0xc4 },	/* SWE2 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 }	/* UK   */
};

static BYTE ISOextens2[][5] = {		/* uppercase/lowercase */
    { 0x2b,0x7c,0x7c,0x26,0x5c },	/* CBM  */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* US   */
    { 0xe4,0xf6,0xfc,0xdf,0x7f },	/* GER  */
    { 0xc6,0xd8,0xc5,0x7e,0x7f },	/* DEN1 */
    { 0xe9,0xf9,0xe8,0xa8,0x7f },	/* FRA  */
    { 0xc4,0xd6,0xc5,0xfc,0x7f },	/* SWE1 */
    { 0xe0,0xf2,0xe8,0xec,0x7f },	/* ITA  */
    { 0xa8,0xf1,0x7d,0x7e,0x7f },	/* SPA  */
    { 0xe6,0xf8,0xe5,0x7e,0x7f },	/* DEN2 */
    { 0xe4,0xf6,0xe5,0xfc,0x7f },	/* SWE2 */
    { 0x7b,0x7c,0x7d,0x7e,0x7f }	/* UK   */
};


/* PC */
static BYTE DOSextens1[][6] = {		/* graphics/uppercase */
    { 0x5b,0x9c,0x5d,0x5e,0x5f,0xc4 },	/* CBM  */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 },	/* US   */
    { 0x8e,0x99,0x9a,0xe1,0x5f,0xc4 },	/* GER  */
    { 0x91,0x6f,0x86,0x5e,0x5f,0xc4 },	/* DEN1 */
    { 0xf8,0x87,0x15,0x5e,0x5f,0xc4 },	/* FRA  */
    { 0x84,0x94,0x86,0x5e,0x5f,0xc4 },	/* SWE1 */
    { 0xf8,0x5c,0x82,0x5e,0x5f,0xc4 },	/* ITA  */
    { 0xad,0xa5,0xa8,0x5e,0x5f,0xc4 },	/* SPA  */
    { 0x92,0x4f,0x8f,0x5e,0x5f,0xc4 },	/* DEN2 */
    { 0x8e,0x99,0x8f,0x5e,0x5f,0xc4 },	/* SWE2 */
    { 0x5b,0x5c,0x5d,0x5e,0x5f,0xc4 }	/* UK   */
};

static BYTE DOSextens2[][5] = {		/* uppercase/lowercase */
    { 0x2b,0x7c,0x7c,0x26,0x5c },	/* CBM  */
    { 0x7b,0x7c,0x7d,0x7e,0x7f },	/* US   */
    { 0x84,0x94,0x81,0x7e,0x7f },	/* GER  */
    { 0x92,0x4f,0x8f,0x7e,0x7f },	/* DEN1 */
    { 0x82,0x97,0x8a,0x20,0x7f },	/* FRA  */
    { 0x8e,0x99,0x8f,0x81,0x7f },	/* SWE1 */
    { 0x85,0x95,0x8a,0x8d,0x7f },	/* ITA  */
    { 0x20,0xa4,0x7d,0x7e,0x7f },	/* SPA  */
    { 0x91,0x6f,0x86,0x5e,0x7f },	/* DEN2 */
    { 0x84,0x94,0x86,0x81,0x7f },	/* SWE2 */
    { 0x7b,0x7c,0x7d,0x7e,0x7f }	/* UK   */
};


/*
 * C64 Uppercase 97-122 (61-7A) and 160-192 (A0-C0) to ASCII
 */

static BYTE ASCgraphics1[] = {
    35,124, 45, 45, 45, 45,124,124, 92, 92, 47, 92, 92, 47, 47, 92,
    35, 95, 35,124, 47, 88, 79, 35,124, 35
};
static BYTE ASCgraphics2[] = {
    32,124, 35, 45, 45,124, 35,124, 35, 47,124,124, 47, 92, 92, 45,
    47, 45, 45,124,124,124,124, 45, 45, 45, 47, 92, 92, 47, 47, 35
};


/* ------------------------------------------------------------------------- */


static BYTE **extens1  = (BYTE **)ASCextens1;	/* defaults for safety */
static BYTE **extens2  = (BYTE **)ASCextens2;
static BYTE *graphics1 = NULL;
static BYTE *graphics2 = NULL;


int  set_locale(char *localename)
{
    if (!localename)
	return (-1);

    if (!strcmp(localename, "Latin1") ||
	!strcmp(localename, "iso_8859_1") ||
	!strcmp(localename, "iso-8859-1")) {

	extens1 = (BYTE **)ISOextens1;
	extens2 = (BYTE **)ISOextens2;
	graphics1 = NULL;
	graphics2 = NULL;
    }
    else {
	extens1 = (BYTE **)ASCextens1;
	extens2 = (BYTE **)ASCextens2;
	graphics1 = NULL;
	graphics2 = NULL;
    }

    return (0);
}


/*
 * PETSCII to ASCII or 8-bit conversion. Alas, there is not just
 * one or two conversions, but a complete chaos...
 *  -- and then there are all the various standard sets...
 */

/*
 * Convert a national CBM character to its standard
 * international equivalent.
 *
 * Parameters:
 *  lang  0-10  CBM language codes
 *  case  0     lowercase, otherwise uppercase
 *     c        byte to convert
 */


int  p_toichar(int lang, int lwcase, BYTE c)
{
    if (lang < 0 || lang > 10) {
	fprintf (stderr, "Fatal error: language not recognized.\n");
	exit (2);
    }

    if ((c >= 0xc0) && (c <= 0xdf))  /* same as 96 - 127 */
	c ^= 0xa0;
    else if ((c >= 0x20) && (c <= 0x40))
	return (c);

    if ((c >= 0x41) && (c <= 0x5a)) {
	if (lwcase)
	    c ^= 0x20;
    }
    else if ((c >= 0x5b) && (c <= 0x60)) {
	if (!lwcase && (c <= 0x5d) &&
	    (lang == CS_DEN1 || lang == CS_SWE1))
	    c = extens2[lang][c -0x5b];
	else
	    c = extens1[lang][c -0x5b];
    }
    else if ((c >= 0x61) && (c <= 0x7a)) {
	if (lwcase)
	    c ^= 0x20;
	else
	    c = graphics1[c -0x61];
    }
    else if ((c >= 0x7b) && (c <= 0x7f)) {
	c = extens2[lang][c -0x7b];
    }

    else if (c >= 0xa0 && c <= 0xfe)
	c = graphics2[c & 0x1f];
    else if (c == 0xff)
	c = 0x7e;

#if 0

    else {
	/* 00-31, 128-159, 160-192, 224-254 */
	if (CLEAN) {
	    /* these codes must be skipped when clean is set */
	    if (c < 0x0a || c == 0xb ||
		(c > 0x0d && c < 0x20) || c > 0x7e)
		return SERIAL_OK;
	}
	else {
	    if ((c < 0x20) && *ctrl1[c])
		return ((fprintf(printer->FileDs, "(%s)",
				 ctrl1[c]) < 0) ?
			SERIAL_ERROR : SERIAL_OK);

	    if (((c & 0x7f) < 0x20) && *ctrl2[c & 0x1f])
		return ((fprintf(printer->FileDs, "(%s)",
				 ctrl2[c & 0x1f]) < 0) ?
			SERIAL_ERROR : SERIAL_OK);
	    return ((fprintf(printer->FileDs, "($%02x)",
			     c & 0xff) < 0) ? SERIAL_ERROR : SERIAL_OK);
	} /* !clean */
    }
#endif

    return (c);
}


/* ------------------------------------------------------------------------- */

/*
 * Language name to code.
 */

int  str_to_lang(char *name)
{
    int     j, token;
    char   *p, *q;

    if (!name || !*name)
	return (-1);

    for (p = name; *p; p++) {
	if (isalpha (*p))
	    *p = tolower(*p);
	else if (*p == '_' || *p == '/')
	    *p = '-';
    }

    for (token = 0; token < NUM_LANGUAGES; token++) {
	for (p = langnames[token], q = (char *)name, j = 0;
	     *p && *q && *p == *q; p++, q++, j++);

	/* found an exact match */
	if (j && !*p) {
	    /*fprintf (stderr, "found %s %2x\n", langnames[token], token);*/
	    return token;
	}
    } /* for */

    return (-1);
}
