/*
 * $Id: config.h,v 1.10 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 */

/* ------------------------------------------------------------------------- *
 *
 * X64 Configuration Set Up
 * Recommendations for each define are shown after them, so
 * that you can restore the settings.
 *
 * Use 'xmkmf' to update the Makefile.
 *
 * ------------------------------------------------------------------------- */

/*
 * Interfaces, Windows
 */


/*
 * X11 front-end for the debugger
 *
 * 'X022' selects the old debugger from X64-0.2.2
 *   (not recommended if the new 'xdebug.c' compiles.)
 */

#define X022			/* no */

#ifndef X022
#ifndef NEW_VIC

#define XDEBUGGER		/* yes */
#endif
#endif


/*
 * If you want to use command line editor with monitor, define following
 * It uses termios interface for a tty by default, uncomment next line
 * if you only have termio
 */

#define EDIT			/* yes */


/*
 * Which way to copy files ?
 * Since C1541 disk drive copies files to the first filename, whereas
 * they normally go to the last file, one can change the 'c1541' copy
 * syntax via this option.
 *   UNIX:  copy    oldname [oldname2 ...] newname
 *   1541:  copy    newname oldname [oldname2 ...]
 */

/*#define COPY_TO_LEFT*/	/* no */


/* ------------------------------------------------------------------------- */

/*
 * Keyboard
 */


/*
 * In case of problems with keyboard you may enable keyboard debug mode
 */

/*#define KBDBUG*/	/* no */	/* full */
/*#define KBDBUG2*/	/* no */	/* keypress stack only */


/*
 * Set up Keyboard Input Buffer in X-version
 *
 * Some keyboards do not generate KeyRelease Event, some have
 * different layout. Try which settings are best for your keyboard.
 */

/*#define AUTOREPEAT*/	/* no */	/* operate without keyrelease events */
/*#define US_KBD*/	/* US only */	/* swap asciicircum with asciitilde */


/*
 * Host keyboard repeat treshold (ms). This filters out any excessive
 * key events, so it must be longer than actual autorepeat interval.
 */

#ifndef C128
#define REPEAT_DELAY  50	/* near 50 */
#else
#define REPEAT_DELAY  100
#endif


/*
 * Some keys only for DEC with additonal Editing Keypad
 */

/*#define DECMAP*/		/* no */


/* ------------------------------------------------------------------------- */

/*
 * CPU
 */


/*
 * define either PAL or NTSC (it does't affect much though)
 */

#define PAL			/* yes */
/*#define NTSC*/		/* no */


/*
 * "Hacked" and faster CPU emulation
 * For CPU debugging, NORMAL (preferred) or HACK2 should be used.
 */

#define HACK2			/* yes */
/*#define HACK*/		/* no */
/*#define NORMAL*/		/* no */


/*
 * If you do NOT want to emulate undocumented commands, uncomment
 * the line below, and also select "NORMAL" CPU emulation.
 */

/*#define NO_UNDOC_CMDS*/	/* no */


/*
 * Uncommenting following line causes some debug information to be printed
 * and with 'NORMAL', different 6510 intructions can be counted.
 */

/*#define DEBUG*/		/* no */


/* ------------------------------------------------------------------------- */

/*
 * Peripherals
 */


/*
 * Joystick (needs the joystick module for Linux)
 */

/*#define JOYSTICK*/		/* Linux only */


/*
 * In case of problems with the disk drive
 */

/*#define DEBUG_1541*/		/* no */

/*#define DEBUG_SERIAL*/	/* no */	/* Serial control */
/*#define DEBUG_SERIAL_1*/	/* no */	/* Serial data flow */


/*
 * Printer
 */

/* Select Default Character Set */

#define PRN_CS_DEFAULT  0	/* 0 */


/* ------------------------------------------------------------------------- */

/*
 * Miscellaneous
 */

/*
 * Shall Virtual RAM Expander be installed ?
 * Sizes available are 128, 256 and 512 KB
 * (C128 internal RAM may be expanded in vmachine.h)
 */

#define REU			/* yes */
#define REUSIZE 256


/*
 * Utility to patch between ROM revisions.
 * Supports 0, 3, 67 and 100 for the C64, but user can define more.
 */

#define PATCH_ROM		/* yes */


/*
 * Emulator Identification Utility
 */

#define EMULATOR_ID		/* yes */


/*
 * Trap idle
 */

/*#define IDLE_TRAP*/		/* no */


/*
 * Experimental timer code
 */

#define NEW_TIMER_CODE		/* yes */


/* ------------------------------------------------------------------------- */

/*
 * Define a few default filenames
 */


#ifndef PROJECTDIR
#ifdef C128
#define PROJECTDIR	"C128"
#else
#ifndef VIC20
#define PROJECTDIR	"C64"
#else
#define PROJECTDIR	"VIC"
#endif
#endif
#endif


#ifndef RAMNAME
#define RAMNAME		"ram"
#endif

#ifndef BASICNAME
#define BASICNAME	"basic"
#endif

#ifndef KERNELNAME
#define KERNELNAME	"kernel"
#endif

#ifndef CHARGENNAME
#define CHARGENNAME	"chargen"
#endif


#ifndef REUNAME
#define REUNAME		"reu"
#endif

#ifndef EXROMNAME
#define EXROMNAME	"exrom"
#endif

#ifdef HPUX
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#endif
