/*
 * $Id: drive.h,v 1.4 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Definitions for both built-in and standalone drive simulators and
 * emulators.
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *   Dan Miner (dminer@nyx10.cs.du.edu)
 *
 * $Log: drive.h,v $
 * Revision 1.4  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Prepared for Error Data handling.
 *
 * Revision 1.3  1995/01/26  13:51:18  jopi
 * Structures and control values for multiple drive handling.
 *
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * Added multi-disk support
 *
 * Revision 1.1  1994/08/10  17:46:41  jopi
 * Initial revision
 *
 *
 *
 */

#ifndef X64_DRIVE_H
#define X64_DRIVE_H


#include <time.h>

#include "types.h"
#include "file.h"


#define UPPER(ad)		(((ad)>>8)&0xff)
#define LOWER(ad)		((ad)&0xff)

#define DRIVE_RAMSIZE		0x400
#define IP_MAX_COMMAND_LEN	128   /* real 58 */


typedef struct bufferinfo_s {
    int     mode;	/* mode on this buffer */
    int     readmode;	/* is this for reading or writing */
    int     secondary;	/* this is not needed */
    BYTE   *buffer;	/* use this to save data */
    BYTE   *slot;	/* save data for directory-slot */
    int     bufptr;	/* use this to save/read data to disk */
    int     length;	/* directory-read length */
}       bufferinfo_t;


/*
 *  Run-time data struct for each drive
 */

typedef struct {
    int     type;		/* Device */

    /* Current image file */

    int     mode;		/* Read/Write */
    int     ActiveFd;
    char    ActiveName[256];	/* Image name */
    time_t  mtime;		/* Write Locking information */
    char    changed;

    /* Drive information */

    int     NumTracks;
    int     NumBlocks;		/* Total Count (683) */
    int     ErrFlg;		/* Flag if Error Data is available */
    char   *ErrData;

    BYTE   *dosrom;
    BYTE   *dosram;
    BYTE    bam[256];
    bufferinfo_t buffers[16];


    /* File information */

    BYTE    find_buffer[256];

    int     find_slot;
    int     Curr_track;
    int     Curr_sector;

    char   *find_name;
    int     find_length;
    int     find_type;
} DRIVE;


/*
 * Actually, serial-code errors ...
 */

#define FLOPPY_COMMAND_OK		0
#define FLOPPY_ERROR			(2)
#define FLOPPY_EOF			(64)


/*
 * Return values used around.
 */

#define FD_OK		0
#define FD_EXIT		1	/* -1,0, 1 are fixed values */

#define FD_NOTREADY	-2
#define FD_NOTRD	-3
#define FD_NOTWRT	-4
#define FD_WRTERR	-5
#define FD_RDERR	-6
#define FD_BADIMAGE	-7	/* ID mismatch (Disk or tape) */
#define FD_BADVAL	-8	/* Illegal value */
#define FD_BADDEV	-9
#define FD_BAD_TS	-10	/* Track or sector */


#define CHK_NUM		0
#define CHK_RDY		1
#define CHK_EMU		2	/* Is image block based */


extern int  attach_floppy_image ( DRIVE *floppy, char *name, int mode );
extern void detach_floppy_image ( DRIVE *floppy );

#endif  /* X64_DRIVE_H */
