/*
 * $Id: extern.h,v 1.9 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 * $Log: extern.h,v $
 * Revision 1.9  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Linux joystick.
 * VIC drawing routines renamed appropriately.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * *** empty log message ***
 *
 * Revision 1.7  1994/08/10  18:34:13  jopi
 * More changeability
 *
 * Revision 1.6  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.5  1994/02/18  16:10:57  jopi
 * New debugger & arg parser, x20 and x128
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:30  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#ifndef X64_EXTERN_H
#define X64_EXTERN_H


#include "types.h"	/* for BYTE, ADDRESS, etc. types and structures */
#include "proto.h"


#ifdef __cplusplus
#ifdef __STDC__
#undef __STDC__
#endif
#define __STDC__ 0
#endif

extern char *progname;
extern int   clength[];

extern BYTE  accumulator;
extern BYTE  x_register;
extern BYTE  y_register;
extern BYTE  stack_pointer;
extern BYTE  status_register;
extern ADDRESS program_counter;
extern CLOCK clk;

extern int   zero_flag;
extern int   sign_flag;
extern int   overflow_flag;
extern int   break_flag;
extern int   decimal_flag;
extern int   interrupt_flag;
extern int   carry_flag;

extern AppResources app_resources;


/* Debugging */
extern int   hexflg;
extern int   verflg;
extern int   debugflg;
extern int   runflg;

extern int   autodump;
extern int   reuflg;


/* Memory */

extern int module;

extern BYTE   ram[];
extern BYTE   kernal_rom[];
extern BYTE   basic_rom[];
extern BYTE   char_rom[];
extern BYTE  *external_rom;
extern BYTE   sid[];
extern BYTE   cia1[];
extern BYTE   cia2[];
extern BYTE   keyarr[];
extern BYTE   joy1, joy2;

extern void  (*(memwr[])) ( ADDRESS, BYTE );
extern BYTE  (*(memrd[])) ( ADDRESS );


/* Screen */
extern BYTE  *charbase;
extern BYTE  *scrbase;
extern BYTE  *bitmap1_base;
extern BYTE  *bitmap2_base;

extern void  (*vic_draw_bitmap) ( ADDRESS );

/* temp */

#ifdef NEW_VIC
extern void   x_redraw ( void );
#else
extern void  (*vic_redraw) ( void );
#endif


/* Keyboard & joystick */

#ifdef JOYSTICK
extern void joystick(void) ;
extern void joyini(void) ;
extern void joyclose(void);

extern int joyfd ;
#endif


/* C64 performance */

extern int rd[];
extern int wr[];
extern long instructions[];


/* Timer */

extern CLOCK int_clk[];

extern CLOCK next_int_clk;
extern int next_int;

#ifdef NEW_TIMER_CODE
/* experimental timer code */

extern CLOCK next_alarm_clk;

#endif /* NEW_TIMER_CODE */


#ifndef SYSV
#if defined (sun) && defined (__GNUC__)
int printf(),toupper(),fprintf(),fwrite(),fread(),fputc(),fgetc(),perror();
int fclose(),strtol(),rewind(),sscanf(),tolower(),remove(),fflush();
#endif
#endif

#endif  /* X64_EXTERN_H */

