/*
 * $Id: main.c,v 1.9 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Main builds up the emulation and controls its operation.
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *
 *
 * $Log: main.c,v $
 * Revision 1.9  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Options and control for Printer, Serial and ROM patcher updated.
 * Signal calls end64 for decent exit.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * *** empty log message ***
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser, x20 and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 * load file reorganised
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:38:36  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>

#include "macro.h"
#include "vmachine.h"
#include "memory.h"
#include "serial.h"
#include "extern.h"
#include "proto.h"
#include "patchlevel.h"

extern int     load_mem ( void );


#ifdef C128
#define EMULATOR  "128"
#else
#ifndef VIC20
#define EMULATOR  "64"
#else
#define EMULATOR  "VIC 20"
#endif
#endif

char *progname;

extern CLOCK   clk;
extern CLOCK   next_int_clk;
extern CLOCK   int_clk[NUMOFINT];
extern int     next_int;

#if 1
int     hexflg = 1;
int     module = 0;
int     verflg = 0;  /* VIC */
int     runflg = 0;  /* ==> app_resources */
int     debugflg = 0;  /* CPU  + CIA */
#endif

int     autodump = 0;
int     reuflg = 0;

#ifdef NEW_TIMER_CODE

extern int new_alarm_number;
extern void new_alarm_trap(int i);
extern void update_intr_alarm(void);

#endif /* NEW_TIMER_CODE */


 /* Code */


int     main(int argc, char **argv)
{
    ADDRESS start_addr;
    int     i;
    int     rl = -1;
    int     len=  0;


    /* Initialisation. */

    app_resources.notraps = 0;
    app_resources.PrinterLang = app_resources.Locale = NULL;

    progname = *argv;
    arg_parser(argc, argv);
    x_start();

    signal(SIGINT, end64);


    /*set_locale(app_resources.Locale);*/

    printf ("\nWelcome to %s, the Commodore %s Emulator for XWindows\n",
	    progname, EMULATOR);
    printf ("Version %3.1f PL %d  %s\n\n", 
	    (float)VERSION, (int)PATCHLEVEL, PATCHMESG);

    if (load_mem() < 0 && !app_resources.asmflag)
	exit (1);

    printf("\nInstalling configuration dependent code...\n\n");


#ifdef PATCH_ROM
#ifdef C128
    if (app_resources.basicRev) {
	(void)patch_rom (app_resources.basicRev);
    }
#endif
    if (app_resources.kernelRev) {
	(void)patch_rom (app_resources.kernelRev);
    }
#endif  /* PATCH_ROM */

#ifdef REU
    if (app_resources.reuName) {
	if (reset_reuc(app_resources.reuName, (REUSIZE << 10)) >= 0)
	    ++reuflg;	/* REU active */
    }
#endif

#ifdef JOYSTICK
    joyini();
#endif

    for (i = 0; i < NUMOFINT; i++)
	int_clk[i] = 0;

    int_clk[I_SPECX] = 1;

    if (!app_resources.notraps) {
	printf("Init Serial Bus.\n");

	initialize_traps();
	initialize_serial();

	initialize_printer(4, app_resources.PrinterLang,
			   app_resources.Locale );

	initialize_1541(8, DT_DISK | DT_1541);
	initialize_1541(9, DT_DISK | DT_1541);
	initialize_1541(11, DT_FS | DT_1541);


	if (app_resources.floppyName)
	    if ( serial_select_file(DT_DISK | DT_1541, 8, app_resources.floppyName) < 0)
		fprintf (stderr, "\nFloppy attach failed.\n");
    } /* Traps */


    /* define start address */
    start_addr = intr(I_RESET, 0);

    if (app_resources.start_addr /*&& *app_resources.start_addr*/)
	start_addr = (ADDRESS) strtol(app_resources.start_addr,
				      NULL, hexflg ? 16 : 10);

    /* Load given program */
    if (app_resources.reloc)
	rl = (ADDRESS) strtol(app_resources.reloc, NULL, hexflg ? 16 : 10);
    if (app_resources.programName) {
	len = LoadFile (NULL, app_resources.programName, ram, rl, RAMSIZE);
	if (rl >= 0)
	    set_text_pointers(rl, rl + len);
    }

    if (app_resources.asmflag)
	mon(start_addr);


#ifdef NEW_TIMER_CODE
    new_alarm_number = new_alarm();
    update_intr_alarm();
#endif /* NEW_TIMER_CODE */

    printf("\nStarting Commodore %s emulator with start address %04X\n",
	   EMULATOR, start_addr);

    mainloop(start_addr);
    printf("perkele!\n");
    exit(0);   /* never reached */
}


void    end64(void)
{
    printf("\n");
    x_exit();

    if (!app_resources.notraps)
	remove_serial(-1);

#ifdef JOYSTICK
    if (joyfd >= 0)
	joyclose();
#endif

#ifdef REU
    if (reuflg)
	close_reuc(app_resources.reuName);
#endif

    if (autodump)
      dump (create_name(NULL, RAMNAME));

#ifdef DEBUG
    /*
     * This routine stores cumulated count of different instructions
     * if profiler is enabled.
     */

    /* if (debugflg) */
    {
	int     i;
	long    total = 0;
	char   *bufp;
	FILE   *fp;

	bufp = create_name(NULL, "inst");
	if (NULL == (fp = fopen(bufp, "w")))
	    return;

	for (i = 0; i < 256; i++) {
	    total += instructions[i];
	    fprintf(fp, "%02X %-4s %8ld\n", i, lookup[i].mnemonic,
		instructions[i]);
	}
	fprintf(fp, "----------------\n\tTotal %8ld\n", total);
	fclose(fp);

	/* Store C64 memory access log file */
	/* if ((bufp = create_name(NULL, "mem")) == NULL ||
	 *   ( fp = fopen( "mem", "w" )) == NULL) return 1;
	 * for ( i=0; i<65536; i++ )
	 *   fprintf( fp, "%04X %10d %10d\n", i, rd[i], wr[i]); fclose( fp );
	 */
    }
#endif

    exit(0);
}
