/*
 * $Id: printer.c,v 1.4 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Restricted Epson compatible Serial printer emulator.
 * Only text mode is supported.
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 *
 * $Log: printer.c,v $
 * Revision 1.4  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Separate Charset Converter module.
 * New serial device control.
 * Header file.
 *
 * Revision 1.3  1994/12/12  16:59:44  jopi
 * Dynamical printer structure for multiple device support.
 *
 * Revision 1.2  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.1  1993/11/10  01:39:37  jopi
 * Initial revision
 * Star LC10 CCL emulator; removes escape sequences.
 * Character conversion from petcat.c
 *
 *
 */


#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <ctype.h>
#include <stdlib.h>
#include <assert.h>

#include "config.h"
#include "types.h"
#include "serial.h"
#include "printer.h"
/*#include "charsets.h"*/
#include "proto.h"


#ifndef __CEXTRACT__

extern int  initialize_printer ( int dev, char *lang, char *locale );
extern void detach_print ( PRINTER *printer );
extern int  attach_print ( PRINTER *printer, char *name, int mode );

extern int  open_printer ( void *printer, char *name, int length, int secondary );
extern int  close_printer ( void *printer, int secondary );
extern int  read_printer ( void *printer, BYTE *data, int secondary );
extern int  write_printer ( void *printer, BYTE data, int secondary );
extern void flush_printer ( void *printer, int secondary );

#endif /* __CEXTRACT__ */


 /*
  * These codes are the same as in petcat.c
  */

char   *ctrl1[] = {
    "", "", "", "", "", "wht", "", "",
    "dish", "ensh", "\n", "", "\f", "\n", "swlc", "",
    "", "down", "rvon", "home", "del", "", "", "",
    "", "", "", "esc", "red", "rght", "grn", "blu"
};

char   *ctrl2[] = {
    "", "orng", "", "", "", "F1", "F3", "F5",
    "F7", "F2", "F4", "F6", "F8", "sret", "swuc", "",
    "blk", "up", "rvof", "clr", "inst", "brn", "lred", "gry1",
    "gry2", "lgrn", "lblu", "gry3", "pur", "left", "yel", "cyn"
};


/*
 * C64 Uppercase 97-122 (61-7A) and 160-192 (A0-C0) to ASCII
 */

static BYTE graphics1[] = {
    35,124, 45, 45, 45, 45,124,124, 92, 92, 47, 92, 92, 47, 47, 92,
    35, 95, 35,124, 47, 88, 79, 35,124, 35
};
static BYTE graphics2[] = {
    32,124, 35, 45, 45,124, 35,124, 35, 47,124,124, 47, 92, 92, 45,
    47, 45, 45,124,124,124,124, 45, 45, 45, 47, 92, 92, 47, 47, 35
};


static BYTE extens1[] = {   /* uppercase/lowercase */
    0x5b,0x9c,0x5d,0x5e,0x5f,0xc4   /* CBM */
#ifdef CS_SWE_
    0x7b,0x7c,0x7d,0x5e,0x5f,0xc4   /* SWE ascii */
#endif
#ifdef CS_GER_
    0x7b,0x7c,0x75,0x5e,0x5f,0xc4   /* GER ascii */
#endif
};

static BYTE extens2[] = {   /* graphics/uppercase */
    0x2b,0x7c,0x7c,0x26,0x5c,

    0x2b,0x7c,0x7c,0x26,0x5c   /* CBM */
#ifdef CS_SWE_
    0x5b,0x5c,0x5d,0x7e,0x23   /* SWE ascii */
#endif
#ifdef CS_GER_
    0x5b,0x5c,0x55,0x7e,0x23   /* GER ascii */
#endif
};


int     initialize_printer(int dev, char *lang, char *locale)
{
    PRINTER *printer;

    /*
     * Create instance of the printer
     */

    printer = (PRINTER *)malloc(sizeof(PRINTER));
    assert(printer);
    memset (printer, 0, sizeof(PRINTER));	/* init all pointers */

    printer->type = DT_PRINTER;
    printer->mgn_right = 80;
    printer->mgn_bottom= 64;

    sprintf(printer->ActiveName, "/tmp/hmm_%d", dev);	/* default */
    /*printer->lp_command = ;*/

    if (!lang || (printer->lang = str_to_lang(lang)) < 0)
	printer->lang = PRN_CS_DEFAULT;

    if (attach_serial_device(dev, (char *)printer, "Printer", read_printer,
	    write_printer, open_printer, close_printer, flush_printer)) {
	printf("Could not initialize printer #%d ????\n", dev);
	return(-1);
    }
    else {
	printf("Printer #%d ready.\n", dev);
    }
    return 0;
}


/* ------------------------------------------------------------------------- */

/*
 * Functions to attach blank files for printing.
 */

void    detach_print(PRINTER *printer)
{
    if (!printer)
	return;

    close_printer((void *)printer, 0);

    if (printer->FileDs != NULL) {
	printf("Detaching printer file %s\n", printer->ActiveName);
    }
}


int     attach_print(PRINTER *printer, char *name, int mode)
{
    detach_print(printer);

    if (!name) {
	printf("No name, detaching image.\n");
	return (-1);
    }

    if ((printer->FileDs = fopen (name, "w")) == NULL) {
	perror (name);
	return (-1);
    }

    strcpy(printer->ActiveName, name);
    printf("Printing to file '%s'.\n", name);

    return (0);
}


/* ------------------------------------------------------------------------- */

int     open_printer(void *prn, char *name, int length, int secondary)
{
    PRINTER *printer = (PRINTER *)prn;

    printer->mode &= ~T_CASE;
    printer->mode |= ((secondary == 7) ? T_LOWC : T_UPPC);

    /*if (printer->ActiveName)
	remove (printer->ActiveName);
    else
	sprintf(printer->ActiveName, "/tmp/hmm");*/

    if ((printer->FileDs = fopen (printer->ActiveName, "wb")) == NULL) {
	fprintf(stderr, "Printer: Cannot create file '%s'\n",
		printer->ActiveName);
	perror (printer->ActiveName);
	return (SERIAL_ERROR);
    }

    return (SERIAL_OK);
}


int     close_printer(void *prn, int secondary)
{
    PRINTER *printer = (PRINTER *)prn;
    char buf[256];

    if (printer->FileDs) {
	fclose (printer->FileDs);
/*
	if (printer->lpCommand) {
	    sprintf (buf, "%s %s", printer->lpCommand, printer->ActiveName);
	    system (buf);
	}
*/
	printer->FileDs = NULL;
    }

/*    if (printer->ActiveName)
	remove (printer->ActiveName);
*/
    return SERIAL_OK;
}


int     read_printer(void *printer, unsigned char *data, int secondary)
{
    return SERIAL_EOF;
}


int     write_printer(void *prn, unsigned char data, int secondary)
{
    PRINTER *printer = (PRINTER *)prn;

    /*
     * In order to print the text it must be cleaned and
     * converted into ASCII
     */

  /* Hex dump not used yet */
    if (printer->mode & T_HEXDUMP)
	return ((fprintf(printer->FileDs, "%02x ", data & 0xff) < 0)
		? SERIAL_ERROR : SERIAL_OK);

    if (printer->skip > 0) {
	--(printer->skip);
	return SERIAL_OK;
    }

    if (!(printer->mode & T_ESC))
	switch (data) {
	  case 0x10: /* Set TABs - 2 parameters */
	    if (!(printer->quote))
		printer->skip = 2;
	    return SERIAL_OK;

	  case 0x0a: /* LF */
	    data = '\n';
	    break;

	  case 0x0d: /* CRLF */
	    printer->quote = 0;
	    printer->mode &= ~(T_OLUPPC | T_OLLOWC);
	    data = '\n';
	    break;

	  case 0x1b: /* ESC */
	    if (printer->quote) {
		if (printer->mode & T_CLEAN) {
		    /* C128 escape -- two bytes must be skipped to clean */
		    printer->skip = 1;
		    return SERIAL_OK;
		} else {
		    return ((fprintf(printer->FileDs, "(esc)") < 0)
			    ? SERIAL_ERROR : SERIAL_OK);
		}
	    } else {
		/* escape -- printer command is to follow */
		printer->mode |= T_ESC;
		return SERIAL_OK;
	    }
	    break;

	  case 0x22: /* quotes */
	    printer->quote ^= data;
	    break;

	  case 0xff:
	    data = 0x23;
	    break;

	  case 0x11: /* no quote: lowercase for one line; quote: "(down)" */
	  case 0x91: /* no quote: uppercase for one line; quote: "(up)" */
	    if (!(printer->quote)) {
		printer->mode &= ~(T_OLUPPC | T_OLLOWC);
		printer->mode |= ((data == 0x11) ? T_OLLOWC : T_OLUPPC);
		return SERIAL_OK;
	    } /* else fall trough */

	  default:
	    /*
	     * PETSCII to ASCII conversion. Alas, there is not just one or two
	     * conversions, but a complete chaos...
	     */
#if 0
	    p_toichar(0 /* lang */, printer->mode & (T_LOWC | T_OLLOWC), data);
#else
	    if ((data >= 0xc0) && (data <= 0xdf))  /* same as 96-127 */
		data ^= 0xa0;
	    else if ((data >= 0x20) && (data <= 0x40))
		break;

	    if ((data >= 0x41) && (data <= 0x5a)) {
		if (printer->mode & (T_LOWC | T_OLLOWC))
		    data ^= 0x20;
	    }
	    else if ((data >= 0x5b) && (data <= 0x60)) {
		data = extens1[data-0x5b];
	    }
	    else if ((data >= 0x61) && (data <= 0x7a)) {
		if (printer->mode & (T_LOWC | T_OLLOWC))
		    data ^= 0x20;
		else
		    data = graphics1[data-0x61];
	    }
	    else if ((data >= 0x7b) && (data <= 0x7f)) {
		data = extens2[data-0x7b];
	    }

	    else if (data >= 0xa0 && data <= 0xfe)
		data = graphics2[data & 0x1f];
	    else if (data == 0xff)
		data = 0x7e;
	    else {
		/* 00-31, 128-159, 160-192, 224-254 */
		if (printer->mode & T_CLEAN) {
		    /* these codes must be skipped when clean is set */
		    if (data < 0x0a || data == 0xb ||
			(data > 0x0d && data < 0x20) || data > 0x7e)
			return SERIAL_OK;
		}
		else {
		    if ((data < 0x20) && *ctrl1[data])
			return ((fprintf(printer->FileDs, "(%s)",
					 ctrl1[data]) < 0) ?
				SERIAL_ERROR : SERIAL_OK);

		    if (((data & 0x7f) < 0x20) && *ctrl2[data & 0x1f])
			return ((fprintf(printer->FileDs, "(%s)",
					 ctrl2[data & 0x1f]) < 0) ?
				SERIAL_ERROR : SERIAL_OK);
		    return ((fprintf(printer->FileDs, "($%02x)",
				data & 0xff) < 0) ? SERIAL_ERROR : SERIAL_OK);
		} /* !clean */
	    } /* else */
#endif
	} /* switch (data) */

    else {
	/* skip multi-byte escape sequences (i.e. printer commands) */
	printer->mode &= ~T_ESC;

	switch (data) {
	  case 0x10: /* Set TABs - 2 parameters */
	    printer->skip = 2;
	    break;
	  case 0x19: case 0x21: case 0x25:
	  case 0x2d: case 0x33:
	    printer->skip = 1;
	    break;
	  case 0x3a:
	    printer->skip = 3;
	    break;

	  case 0x41: case 0x43: case 0x4a:
	  case 0x4e: case 0x51:
	    printer->skip = 1;
	    break;

	  case 0x52:
	    printer->skip = 1;
	    break;

  	  case 0x53:
	  case 0x57:
	  case 0x5d: /* select operating mode */
	    printer->skip = 1;
	    break;

	  case 0x61: case 0x63: case 0x68:
	  case 0x6b: case 0x6c: case 0x70:
	  case 0x72: case 0x77: case 0x7e:
	    printer->skip = 1;
	    break;

	    /* unsupported codes:
	     * (dec) 18, 38, 42, 43_long, 66, 75, 76, 89, 90
	     */
	}

	return SERIAL_OK;
    } /* esc (else) */

    return fputc (data, printer->FileDs) == EOF ? SERIAL_ERROR : SERIAL_OK;
}


void    flush_printer(void *printer, int secondary)
{
    if (((PRINTER *)printer)->FileDs)
	fflush (((PRINTER *)printer)->FileDs);
    return;
}
