/*
 * $Id: serial.h,v 1.9 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Teemu Rantanen (tvr@cs.hut.fi)
 *
 * $Log: serial.h,v $
 * Revision 1.9  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * New serial device control.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * *** empty log message ***
 *
 * Revision 1.7  1994/08/10  18:34:13  jopi
 * More changeability
 *
 * Revision 1.6  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.5  1994/06/08  02:06:23  jopi
 * Patchlevel 2
 * Serial device types
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:39:24  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#ifndef X64_SERIAL_H
#define X64_SERIAL_H

#include "types.h"

#define SERIAL_OK		0
#define SERIAL_FILE_NOT_FOUND	64
#define SERIAL_ERROR		64
#define SERIAL_EOF		64


/*
 * Device types
 */

#define DT_1541			0
#define DT_1570			1
#define DT_1571			2
#define DT_1581			3

#define DT_MASK			7

#define DT_FS			0x10	/* Filesystem Drive */
#define DT_DISK			0x20
#define DT_PRINTER		0x40



#if defined(__STDC__)
typedef struct serial_s
{
    int inuse;
    int isopen;
    char *info;					/* pointer to run-time data */
    char *name;					/* name of the device */
    int (*getf)(void *, BYTE *, int);		/* serial read function */
    int (*putf)(void *, BYTE, int);		/* serial write function */
    int (*openf)(void *, char *, int, int);	/* serial open function */
    int (*closef)(void *, int);			/* serial close function */
    void (*flushf)(void *, int);		/* tell device that write completed */
} serial_t;

#else
typedef struct serial_s
{
    int inuse;
    int isopen;
    char *info;			/* pointer to run-time data */
    char *name;			/* name of the device */
    int (*getf)();		/* serial read function */
    int (*putf)();		/* serial write function */
    int (*openf)();		/* serial open function */
    int (*closef)();		/* serial close function */
    void (*flushf)();		/* tell device that write completed */
} serial_t;

#endif

#endif  /* X64_SERIAL_H */
