/*
 * $Id: sun64.c,v 1.8 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by 
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 *
 * $Log: sun64.c,v $
 * Revision 1.8  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * VIC drawing routines renamed appropriately.
 *
 * Revision 1.7  1994/12/12  16:59:44  jopi
 * Compability
 *
 * Revision 1.6  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:39:27  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

/* 
 *        *** IMPORTANT ***
 *
 *        for sun users only
 *
 * this file is bad for your health, unless you
 * are a hacker of some kind. (And I mean it.)
 *
 * if you still decide to try this, and
 * make it with 'make -f Makefile.sun64',
 * be prepared to restore kbd_mode back to 
 * normal and clear_colormaps etc.
 * because of debug information, you
 * should either direct it to file or
 * run the program from another machine.
 *
 * Help -key quits the emulator
 *
 * runs on SunOS 4.1.2, 4.1.3 and 5.2
 * on 5.3 you have to log on the target
 * machine too... _WITHOUT_ X !
 *
 *   Have fun!
 *
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <memory.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>

#ifndef SYSV
#include <sun/fbio.h>
#include <sundev/kbio.h>
#include <sundev/kbd.h>
#include <sundev/vuid_event.h>
#else
#include <sys/fbio.h>
#include <sys/kbio.h>
#include <sys/kbd.h>
#include <sys/vuid_event.h>
#endif

#ifdef BF
#undef BF
#endif

#include "macro.h"
#include "vmachine.h"
#include "extern.h"

void    (*vic_draw_square) ( ADDRESS );
void    (*vic_draw_bitmap) ( ADDRESS );
void    (*vic_redraw) ( void );

void    vic_set_vmode ( int mode );
void    redraw_text ( void );
/*void    redraw_blank( void );*/

void    draw_nop ( void );
void    flush_queue ( void );

void    aargh ( ADDRESS address );
void    syserr ( char *msg );

/*
void    open_win();
void    open_cntl_win();

void    blurb();
void    open_cntl_win();
void    redraw_cntl_win();
void    toggle_cntl_win();
void    update_cntl_win();
void    update_asm();
int     draw_string();
void    flush_queue();
void    get_colors();
*/


#define START	262144
#define END	1298990
#define MAX_COLORS 16
#define FIRST_COLOR 2
#define XSIZE	40
#define YSIZE	25

int     width, height;
int     xor, yor;
int     xsize = 40 * 8;
int     ysize = 25 * 8;
int     pixs;
int     keypressed;
BYTE	joy1, joy2;

int     kfd, fd;
int     oldtrans;


typedef struct {
    u_short id;
    BYTE    keyarr[8];
}       keyconv;

/* only one key simultaneously is supported */

keyconv keyconvmap[] = {
    {0x08, 1, 0, 0, 0, 0, 0, 0, 0},
    {0x0d, 2, 0, 0, 0, 0, 0, 0, 0},
    {VKEY_PAD6, 4, 0, 0, 0, 0, 0, 0, 0},
    {KEY_TOP(7), 8, 0, 0, 0, 0, 0, 0, 0},
    {KEY_TOP(1), 16, 0, 0, 0, 0, 0, 0, 0},
    {KEY_TOP(3), 32, 0, 0, 0, 0, 0, 0, 0},
    {KEY_TOP(5), 64, 0, 0, 0, 0, 0, 0, 0},
    {VKEY_PAD2, 128, 0, 0, 0, 0, 0, 0, 0},

    {'3', 0, 1, 0, 0, 0, 0, 0, 0},
    {'w', 0, 2, 0, 0, 0, 0, 0, 0},
    {'a', 0, 4, 0, 0, 0, 0, 0, 0},
    {'4', 0, 8, 0, 0, 0, 0, 0, 0},
    {'z', 0, 16, 0, 0, 0, 0, 0, 0},
    {'s', 0, 32, 0, 0, 0, 0, 0, 0},
    {'e', 0, 64, 0, 0, 0, 0, 0, 0},
    {SHIFT_LEFT, 0, 128, 0, 0, 0, 0, 0, 0},

    {'5', 0, 0, 1, 0, 0, 0, 0, 0},
    {'r', 0, 0, 2, 0, 0, 0, 0, 0},
    {'d', 0, 0, 4, 0, 0, 0, 0, 0},
    {'6', 0, 0, 8, 0, 0, 0, 0, 0},
    {'c', 0, 0, 16, 0, 0, 0, 0, 0},
    {'f', 0, 0, 32, 0, 0, 0, 0, 0},
    {'t', 0, 0, 64, 0, 0, 0, 0, 0},
    {'x', 0, 0, 128, 0, 0, 0, 0, 0},

    {'7', 0, 0, 0, 1, 0, 0, 0, 0},
    {'y', 0, 0, 0, 2, 0, 0, 0, 0},
    {'g', 0, 0, 0, 4, 0, 0, 0, 0},
    {'8', 0, 0, 0, 8, 0, 0, 0, 0},
    {'b', 0, 0, 0, 16, 0, 0, 0, 0},
    {'h', 0, 0, 0, 32, 0, 0, 0, 0},
    {'u', 0, 0, 0, 64, 0, 0, 0, 0},
    {'v', 0, 0, 0, 128, 0, 0, 0, 0},

    {'9', 0, 0, 0, 0, 1, 0, 0, 0},
    {'i', 0, 0, 0, 0, 2, 0, 0, 0},
    {'j', 0, 0, 0, 0, 4, 0, 0, 0},
    {'0', 0, 0, 0, 0, 8, 0, 0, 0},
    {'m', 0, 0, 0, 0, 16, 0, 0, 0},
    {'k', 0, 0, 0, 0, 32, 0, 0, 0},
    {'o', 0, 0, 0, 0, 64, 0, 0, 0},
    {'n', 0, 0, 0, 0, 128, 0, 0, 0},

    {'+', 0, 0, 0, 0, 0, 1, 0, 0},
    {'p', 0, 0, 0, 0, 0, 2, 0, 0},
    {'l', 0, 0, 0, 0, 0, 4, 0, 0},
    {'-', 0, 0, 0, 0, 0, 8, 0, 0},
    {'.', 0, 0, 0, 0, 0, 16, 0, 0},
    {':', 0, 0, 0, 0, 0, 32, 0, 0},
    {'@', 0, 0, 0, 0, 0, 64, 0, 0},
    {',', 0, 0, 0, 0, 0, 128, 0, 0},

  /*{ XK_sterling,	0,0,0,0,0,0,1,0 }, */
    {'*', 0, 0, 0, 0, 0, 0, 2, 0},
    {';', 0, 0, 0, 0, 0, 0, 4, 0},
    {VKEY_PAD7, 0, 0, 0, 0, 0, 0, 8, 0},
    {SHIFT_RIGHT, 0, 0, 0, 0, 0, 0, 16, 0},
    {'=', 0, 0, 0, 0, 0, 0, 32, 0},
    {'^', 0, 0, 0, 0, 0, 0, 64, 0},	/* up arrow */
    {'/', 0, 0, 0, 0, 0, 0, 128, 0},

    {'1', 0, 0, 0, 0, 0, 0, 0, 1},
    {'_', 0, 0, 0, 0, 0, 0, 0, 2},	/* right arrow */
    {SHIFT_LEFTCTRL, 0, 0, 0, 0, 0, 0, 0, 4},	/* CNTRL */
    {'2', 0, 0, 0, 0, 0, 0, 0, 8},
    {' ', 0, 0, 0, 0, 0, 0, 0, 16},
    {SHIFT_ALT, 0, 0, 0, 0, 0, 0, 0, 32},  /* CBM */
    {'q', 0, 0, 0, 0, 0, 0, 0, 64},
    {0x1b, 0, 0, 0, 0, 0, 0, 0, 128},	/* STOP */

    /* shifted keys */

    {0x7f, 1, 0, 0, 0, 0, 0, 16, 0},
    {0x0a, 2, 0, 0, 0, 0, 0, 16, 0},
    {VKEY_PAD4, 4, 0, 0, 0, 0, 0, 16, 0},
    {KEY_TOP(8), 8, 0, 0, 0, 0, 0, 16, 0},
    {KEY_TOP(2), 16, 0, 0, 0, 0, 0, 16, 0},
    {KEY_TOP(4), 32, 0, 0, 0, 0, 0, 16, 0},
    {KEY_TOP(6), 64, 0, 0, 0, 0, 0, 16, 0},
    {VKEY_PAD8, 128, 0, 0, 0, 0, 0, 16, 0},

    {'#', 0, 1, 0, 0, 0, 0, 16, 0},
    {'W', 0, 2, 0, 0, 0, 0, 16, 0},
    {'A', 0, 4, 0, 0, 0, 0, 16, 0},
    {'$', 0, 8, 0, 0, 0, 0, 16, 0},
    {'Z', 0, 16, 0, 0, 0, 0, 16, 0},
    {'S', 0, 32, 0, 0, 0, 0, 16, 0},
    {'E', 0, 64, 0, 0, 0, 0, 16, 0},
    /* { XK_Shift_L,	0,128,0,0,0,0,16,0 }, */

    {'%', 0, 0, 1, 0, 0, 0, 16, 0},
    {'R', 0, 0, 2, 0, 0, 0, 16, 0},
    {'D', 0, 0, 4, 0, 0, 0, 16, 0},
    {'&', 0, 0, 8, 0, 0, 0, 16, 0},
    {'C', 0, 0, 16, 0, 0, 0, 16, 0},
    {'F', 0, 0, 32, 0, 0, 0, 16, 0},
    {'T', 0, 0, 64, 0, 0, 0, 16, 0},
    {'X', 0, 0, 128, 0, 0, 0, 16, 0},

    {'\'', 0, 0, 0, 1, 0, 0, 16, 0},
    {'Y', 0, 0, 0, 2, 0, 0, 16, 0},
    {'G', 0, 0, 0, 4, 0, 0, 16, 0},
    {'(', 0, 0, 0, 8, 0, 0, 16, 0},
    {'B', 0, 0, 0, 16, 0, 0, 16, 0},
    {'H', 0, 0, 0, 32, 0, 0, 16, 0},
    {'U', 0, 0, 0, 64, 0, 0, 16, 0},
    {'V', 0, 0, 0, 128, 0, 0, 16, 0},

    {')', 0, 0, 0, 0, 1, 0, 16, 0},
    {'I', 0, 0, 0, 0, 2, 0, 16, 0},
    {'J', 0, 0, 0, 0, 4, 0, 16, 0},
 /* { XK_0, 		0,0,0,0,8,0,16,0 }, */
    {'M', 0, 0, 0, 0, 16, 0, 16, 0},
    {'K', 0, 0, 0, 0, 32, 0, 16, 0},
    {'O', 0, 0, 0, 0, 64, 0, 16, 0},
    {'N', 0, 0, 0, 0, 128, 0, 16, 0},

 /* { XK_plus, 		0,0,0,0,0,1,16,0 }, */
    {'P', 0, 0, 0, 0, 0, 2, 16, 0},
    {'L', 0, 0, 0, 0, 0, 4, 16, 0},
 /* { XK_minus,		0,0,0,0,0,8,16,0 }, */
    {'>', 0, 0, 0, 0, 0, 16, 16, 0},
    {'[', 0, 0, 0, 0, 0, 32, 16, 0},
 /* { XK_at, 		0,0,0,0,0,64,16,0 }, */
    {'<', 0, 0, 0, 0, 0, 128, 16, 0},

    /*
     * { XK_sterling,	0,0,0,0,0,0,17,0 },
     * { XK_asterisk, 0,0,0,0,0,0,18,0
     * },
     */
    {']', 0, 0, 0, 0, 0, 0, 20, 0},
    /*
     * { XK_Home, 		0,0,0,0,0,0,24,0 },
     * { XK_Shift_R, 0,0,0,0,0,0,16,0 },
     * { XK_equal,		0,0,0,0,0,0,48,0 },
     * { XK_asciicircum,	0,0,0,0,0,0,80,0 },
*//* shift up arrow */
    {'?', 0, 0, 0, 0, 0, 0, 144, 0},

    {'!', 0, 0, 0, 0, 0, 0, 16, 1},
    /* { XK_underscore,	0,0,0,0,0,0,16,2 }, *//* right arrow */
    /* { XK_Control_L,	0,0,0,0,0,0,16,4 }, *//* CNTRL */
    {'"', 0, 0, 0, 0, 0, 0, 16, 8},
    /*
     * { XK_space,		0,0,0,0,0,0,16,16 }, { XK_Alt_L,
     * ,0,0,0,0,0,16,32 },
*//* CBM */
    {'Q', 0, 0, 0, 0, 0, 0, 16, 64},
    /* { XK_Escape,	0,0,0,0,0,0,16,128 }, *//* STOP */

};

#define CONV_KEYS (sizeof(keyconvmap)/sizeof(keyconv))


/* ------------------------------------------------------------------------- */

/*
 * Screen
 */

char   *fbbase;
struct fbtype fbuf;

unsigned char oldr[MAX_COLORS + 4];
unsigned char oldg[MAX_COLORS + 4];
unsigned char oldb[MAX_COLORS + 4];
struct fbcmap oldcmap = {
    FIRST_COLOR, MAX_COLORS + 4, oldr, oldg, oldb
};

struct rgb {
    unsigned char red;
    unsigned char green;
    unsigned char blue;
};

unsigned char newr[MAX_COLORS] = {
      0, 255, 255,   0, 160,   0,   0, 255, 
    255, 165, 255,  80, 150, 127, 180, 200
};
unsigned char newg[MAX_COLORS] = {
      0, 255,   0, 255,  32, 255,   0, 255, 
    165,  42, 192,  80, 150, 255, 180, 200
};
unsigned char newb[MAX_COLORS] = {
      0, 255,   0, 255, 240,   0, 200,   0, 
      0,  42, 203,  80, 150, 127, 255, 200
};

struct fbcmap cmap = {
    FIRST_COLOR, MAX_COLORS, newr, newg, newb
};

struct fbcmap bgcolors[20];
struct rgb colorarr[20];

#define SET_RGB(st,co) \
    st.red = newr[co], st.green = newg[co], st.blue = newb[co]


/* ------------------------------------------------------------------------- */

void    x_start(void)
{
    int     i;
    int     keyset;
  /*int     ioctl();*/
    int     select();


    for (i = 0; i < 15; i++) {
	bgcolors[i].index = FIRST_COLOR + MAX_COLORS + i;
	bgcolors[i].count = 1;
	bgcolors[i].red = &(colorarr[i].red);
	bgcolors[i].green = &(colorarr[i].green);
	bgcolors[i].blue = &(colorarr[i].blue);
    }

    if ((kfd = open("/dev/kbd", O_RDWR | O_NDELAY)) < 0)
	syserr("open");

    keyset = 1;
    if (ioctl(kfd, KIOCSDIRECT, &keyset) < 0)
	syserr("kioctl KIOCSDIRECT");

    keyset = 1;
    if (ioctl(kfd, KIOCGTRANS, &oldtrans) < 0)
	syserr("kioctl KIOCGTRANS");

    keyset = TR_EVENT;
    if (ioctl(kfd, KIOCTRANS, &keyset) < 0)
	syserr("kioctl KIOCTRANS");

    if ((fd = open("/dev/fb", O_RDWR)) < 0)
	syserr("open");

    ioctl(fd, FBIOGTYPE, &fbuf);
    width = fbuf.fb_width;
    height = fbuf.fb_height;

    pixs = width / xsize;
    i = height / ysize;
    if (i < pixs)
	pixs = i;
    ysize *= pixs;
    xsize *= pixs;

    xor = (width - xsize) / 2;
    yor = (height - ysize) / 2;

    printf("screen size %dx%d\n", width, height);

    if (ioctl(fd, FBIOGETCMAP, &oldcmap) < 0)
	syserr("ioctl-get");

    if (ioctl(fd, FBIOPUTCMAP, &cmap) < 0)
	syserr("ioctl-cmap");

    fbbase = mmap((caddr_t) 0, width * height, PROT_READ | PROT_WRITE,
	MAP_SHARED, fd, (off_t) START);
    if (fbbase == (caddr_t) - 1)
	syserr("mmap failed");

    printf("size %d\n", fbuf.fb_size);

    {
	register int   i, j;

	/* draw border */
	char   *p, c = FIRST_COLOR + MAX_COLORS + C_BORDER;

	for (p = fbbase, i = yor * width; i; i--, p++)
	    *p = c;

	for (p = fbbase + (yor + ysize) * width,
	    i = (height - ysize - yor) * width; i; i--, p++)
	    *p = c;

	for (i = yor; i < yor + ysize; i++) {
	    for (p = fbbase + i * width, j = xor; j; j--, p++)
		*p = c;
	    for (p = fbbase + i * width + xor + xsize, j = width - xor - xsize;
		j; j--, p++)
		*p = c;
	}
	/* draw background */
	for (i = yor; i < yor + ysize; i++) {
	    for (p = fbbase + i * width + xor, j = xsize; j; j--, p++)
		*p = FIRST_COLOR + MAX_COLORS + C_BG0;
	}
    }

    x_change_color(C_BORDER,1);
    x_change_color(C_BG0,2);

    vic_set_vmode (V_TILT);
    vic_redraw();
}


void    x_loop()
{
    return;
}


void    x_exit()
{
    if (ioctl(fd, FBIOPUTCMAP, &oldcmap) < 0)
	perror("ioctl");
    if (ioctl(kfd, KIOCTRANS, &oldtrans) < 0)
	perror("kioctl");
    close(fd);
    close(kfd);
    printf("exiting...\n");
}


/* --------------------------------------------------------------------------*/

int     x_get_keyarr(void)
{
    struct timeval tm;
    struct firm_event fe;
    int     readm;
    int     i, j, f = 1, k = keypressed;
    int     select();

    while (f) {
	tm.tv_sec = 0;
	tm.tv_usec = 0;
	readm = 1 << kfd;
	if (0 > (f = select(kfd + 1, &readm, NULL, NULL, &tm)))
	    perror("select");
	if (f == 1)
	    if (sizeof (fe) != read(kfd, (char *)&fe, sizeof (fe)))
		perror("read");
	    else {
		if (fe.value)
		    k = fe.id;
		else
		    k = 0;
	    }
    }

    if (k != keypressed) {
	keypressed = k;
	for (j = 0; j < 8; j++)
	    keyarr[j] = 0;
	if (keypressed) {
	    if (keypressed == 32569) { /* Help key */
		x_exit();
		exit (0);
	    }
	    for (i = 0; i < CONV_KEYS; i++) {
		if (keypressed == keyconvmap[i].id) {
		    for (j = 0; j < 8; j++)
			keyarr[j] = keyconvmap[i].keyarr[j];
		    break;
		}
	    }

	     printf ("key down %d %d\n", keypressed, i);
	    if (debugflg) {
		show_keyarr();
	    }
	}
    }
}


/* ------------------------------------------------------------------------- */

void    syserr(msg)
    char   *msg;
{
    extern int errno, sys_nerr;
    extern char *sys_errlist[];

    (void) fprintf(stderr, "ERROR: %s (%d", msg, errno);
    if (errno > 0 && errno < sys_nerr)
	(void) fprintf(stderr, "; %s)\n", sys_errlist[errno]);
    else
	(void) fprintf(stderr, ")\n");
    exit(1);
}


/* ------------------------------------------------------------------------- */

/*
 * draw functions called when square in screen have to be drawn
 */

void    aargh(ADDRESS address)
{
    printf("AAAAARRRRGGHH!!!!! %x \n", (int) address);
}


void    draw_nop(void)
{
    return;
}


void    draw_text(adr)
     ADDRESS adr;
{
    int     x, y, i, j, t, k, l;
    char   *p, *q, *r, col, co;

    adr &= 0x3ff;
    x = adr % 40;
    y = adr / 40;
    col = read_colorram(adr) & 0xf;
    q = (char *) charbase + scrbase[adr] * 8;
    p = (char *) fbbase + (yor + y * 8 * pixs) * width + xor + x * 8 * pixs;

    for (i = 0; i < 8; i++, p += width * pixs)  /* width*pixs = 3456 */
	for (j = 0, t = q[i]; j < 8 * pixs; t = t << 1, j += pixs) {
	    co = FIRST_COLOR + ((t & 0x80) ? col : MAX_COLORS + C_BG0);
	    /* draw one pixel */
	    r = p + j;
	    switch (pixs) {
	      case 1:
		*r = co;
		break;
	      case 2:
		r[width + 1] = r[width] = r[1] = *r = co;
		break;
	      case 3:
		r[2] = r[1] = *r = co;
		r[width + 2] = r[width + 1] = r[width] = co;
		r[2 * width + 2] = r[2 * width + 1] = r[2 * width] = co;
		break;
	      default:
		for (k = 0; k < pixs; k++ , r += width )
		    for (l = 0; l < pixs; l++)
			r[l] = co;
	    }
	}
}


void    draw_mtext(adr)
     ADDRESS adr;
{
    int     x, y, i, j, t, k, l;
    char   *p, *q, *r, c0, c1, c2, c3, co;
    BYTE    c;

    adr &= 0x3ff;
    x = adr % 40;
    y = adr / 40;
    c = scrbase[adr];
    q = (char *) charbase + (c << 3);
    p = (char *) fbbase + (yor + y * 8 * pixs) * width + xor + x * 8 * pixs;
    c0 = FIRST_COLOR + MAX_COLORS + C_BG0;
    c1 = FIRST_COLOR + MAX_COLORS + C_BG1;
    c2 = FIRST_COLOR + MAX_COLORS + C_BG2;
    c3 = FIRST_COLOR + (read_colorram(adr) & 0x7);

    for (i = 0; i < 8; i++, p += width * pixs)
	for (j = 0, t = q[i]; j < 8 * pixs; t = t << 2, j += pixs + pixs) {
	    co = (t & 0xc0) == 0 ? c0 : 
		((t & 0xc0) == 0x40 ? c1 : 
		 ((t & 0xc0) == 0x80 ? c2 : c3));
	    /* draw one pixel */
	    r = p + j;
	    switch (pixs) {
	      case 1:
		r[1] = *r = co;
		break;
	      case 2:
		r[width + 1] = r[width] = r[1] = *r = co;
		r[width + 2] = r[width + 3] = r[2] = r[3] = co;
		break;
	      case 3:
		r[2] = r[1] = *r = co;
		r[5] = r[4] = r[3] = co;
		r[width + 2] = r[width + 1] = r[width] = co;
		r[width + 5] = r[width + 4] = r[width + 3] = co;
		r[2 * width + 2] = r[2 * width + 1] = r[2 * width] = co;
		r[2 * width + 5] = r[2 * width + 4] = r[2 * width + 3] = co;
		break;
	      default:
		for (k = 0; k < pixs; k++ , r += width )
		    for (l = 0; l < pixs; l++)
			r[l +  1] = r[l] = co;
	    }
	}
}

void    draw_ectext(adr)
     ADDRESS adr;
{
    int     x, y, i, j, t, k, l;
    char   *p, *q, *r, col, co, bg;
    BYTE    c;

    adr &= 0x3ff;
    x = adr % 40;
    y = adr / 40;
    c = scrbase[adr];
    q = (char *) charbase + ((c & 0x3f) << 3);
    p = (char *) fbbase + (yor + y * 8 * pixs) * width + xor + x * 8 * pixs;
    col = FIRST_COLOR + (read_colorram(adr) & 0xf);
    switch (c >> 6) {
      case 0:
	bg = FIRST_COLOR + MAX_COLORS + C_BG0;
	break;
      case 1:
	bg = FIRST_COLOR + MAX_COLORS + C_BG1;
	break;
      case 2:
	bg = FIRST_COLOR + MAX_COLORS + C_BG2;
	break;
      case 3:
	bg = FIRST_COLOR + MAX_COLORS + C_BG3;
	break;
      default:
	printf("perrkele\n");
	end64();
    }

    for (i = 0; i < 8; i++, p += width * pixs)
	for (j = 0, t = q[i]; j < 8 * pixs; t = t << 1, j += pixs) {
	    co = (t & 0x80) ? col : bg;
	    /* draw one pixel */
	    r = p + j;
	    switch (pixs) {
	      case 1:
		*r = co;
		break;
	      case 2:
		r[width + 1] = r[width] = r[1] = *r = co;
		break;
	      case 3:
		r[2] = r[1] = *r = co;
		r[width + 2] = r[width + 1] = r[width] = co;
		r[2 * width + 2] = r[2 * width + 1] = r[2 * width] = co;
		break;
	      default:
		for (k = 0; k < pixs; k++ , r += width )
		    for (l = 0; l < pixs; l++)
			r[l] = co;
	    }
	}
}


/* redraws */

void    redraw_tilt()
{
    int     i, j;
    char *p;
    /* draw background */
    for (i = yor; i < yor + ysize; i++) {
	for (p = fbbase + i * width + xor, j = xsize; j; j--, p++)
	    *p = FIRST_COLOR;
    }
}


void    redraw_text()
{
    ADDRESS     i;
    for (i = 0; i < 40 * 25; i++)
	vic_draw_screen(i);
}


/* --------------------------------------------------------------------------*/

void    x_change_color(int i, BYTE color)
{

    SET_RGB( colorarr[i], ((int)(color & 0xf)) );
    if (ioctl(fd, FBIOPUTCMAP, &bgcolors[i]) < 0)
	syserr("ioctl-c");
    
       printf("set color %d: %d (%d %d %d)\n", i, ((int)(color & 0xf)),
	   *bgcolors[i].red, *bgcolors[i].green, *bgcolors[i].blue );
}


void    putpixel(x, y, c)
    int     x, y, c;
{
    int     k, l;
    char   *p;

    p = (char *) fbbase + (yor + y * pixs) * width + xor + x * pixs;

    for (k = 0; k < pixs; k++)
	for (l = 0; l < pixs; l++)
	    p[k * width + l] = c;
}

void	vic_draw_screen(addr)
     ADDRESS addr;
{
    vic_draw_square(addr);
}

void    x_scroll(xscroll, yscroll, oldx, oldy)
     int xscroll, yscroll, oldx, oldy;
{}


void    flush_queue ( void )
{
    return;
};


/* ------------------------------------------------------------------------- */

void    vic_set_vmode(mode)
    int     mode;
{
    printf("vmode set to %d\n", mode);

    switch (mode) {
      case V_TEXT:
        vic_draw_square = draw_text;
/*      vic_draw_bitmap = draw_chargen;*/
        vic_redraw = redraw_text;
        break;

      case V_MTEXT:
        vic_draw_square = draw_mtext;
/*      vic_draw_bitmap = draw_chargen;*/
        vic_redraw = redraw_text;
        break;

      case V_ECTEXT:
        vic_draw_square = draw_ectext;
/*      vic_draw_bitmap = draw_chargen;*/
        vic_redraw = redraw_text;
        break;

/*      case V_BITMAP:
        vic_draw_square = draw_bitmap_square;
        vic_draw_bitmap = draw_bitmap;
        vic_redraw = redraw_bitmap;
        break;

      case V_MBITMAP:
        vic_draw_square = draw_mbitmap_square;
        vic_draw_bitmap = draw_bitmap;
        vic_redraw = redraw_mbitmap;
        break;

      case V_BLANK:
        vic_draw_square = aargh;
        vic_draw_bitmap = aargh;
        vic_redraw = redraw_blank;
        break;
*/

      case V_TILT:
        vic_draw_square = aargh;
        vic_draw_bitmap = aargh;
        vic_redraw = redraw_tilt;
        break;

    }

    (*vic_redraw)();
}

void   x_draw_sprite(i)
    int     i;
{ return; }
void   x_undraw_sprite(i)
    int     i;
{ return; }
void   x_move_sprite(i, x, y)
    int     i, x, y;
{}
void   x_change_sprite(i, q, epx, epy)
    int     i, epx, epy;
    BYTE   *q;
{}
