/*
 * $Id: vdc.c,v 1.1 1995/04/01 07:50:52 jopi Exp $
 *
 * This file is part of the Commodore 128 emulator.
 * See README for copyright notice
 *
 * This file implements the 80-column Video Display Controller.
 * Uses drawing routines from xscreen.c by Jarkko Sonninen (sonninen@lut.fi)
 *
 * Written by
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *
 *
 * $Log: vdc.c,v $
 * Revision 1.1  1995/04/01  07:50:52  jopi
 * Initial revision
 *
 *
 */


#include <stdio.h>
#include <stdlib.h>

#include "macro.h"
#include "extern.h"
#include "vmachine.h"
#include "memory.h"


/* global */

static BYTE   vreg[37];
static BYTE   vdcRAM[16384];	/* possible values: 16K, 2*32K, or 64K */

typedef struct {	/* Pointers to RAM */
    int   mode;
    int   xpix, ypix;
    int   rows, columns;
    BYTE *screen;
    BYTE *color;
    BYTE *chargen;
} VDC;

static VDC vdc;


/*
#define dprintf(a)  printf(a)
*/
#define dprintf(a)

/* functions */

extern int  vdc_init ( void );
extern void store_vdc ( BYTE reg, BYTE byte );
extern BYTE read_vdc ( BYTE reg );


#include <X11/Xlib.h>  /* need Window definition */

void    (*vdc_draw_square) ( VDC *vdc, Window win, ADDRESS );
void    (*vdc_draw_bitmap) ( VDC *vdc, Window win, ADDRESS );
void    (*vdc_redraw) ( void );


/* ------------------------------------------------------------------------- */

/*
 * Interface
 */

void store_vdc(BYTE reg, BYTE byte)
{
    /*if ((address & 1) {*/

    if (reg >= 37)
	return;

    vreg[reg] = byte;

    switch (reg) {
      case 1:		/* Columns displayed */
	vdc.columns = byte;
	return;

      case 6:		/* Rows displayed */
	vdc.rows = byte;
	return;

      case 9:		/* Scan Lines per Character */
	return;

      case 10:	/* Cursor Mode */
      case 11:
	return;

      case 12:	/* Display Address */
      case 13:
	vdc.screen = vdcRAM + (vreg[12] << 8) + vreg[13];
	return;

      case 14:	/* Cursor Address */
      case 15:
	return;

      case 18:	/* Video RAM Address */
      case 19:
	return;

      case 20:	/* Colour Address */
      case 21:
	vdc.color = vdcRAM + (vreg[20] << 8) + vreg[21];
	return;

      case 24:	/* Vertical Scroll */
	return;

      case 25:	/* Display Modes, Horisontal Scroll */
	return;

      case 26:	/* Display Colour */
	return;

      case 28:	/* Char Set Address */
	vdc.chargen = vdcRAM + ((vreg[28] & 0xe0) << 8);
	return;

      case 29:	/* Underline Scan Count */
	return;

      case 30:	/* Character Count (Copy/fill repetition) */
	vreg[0x12] = LOAD(0xa3d);
	vreg[0x13] = LOAD(0xa3c);
	return;

      case 31:	/* Video RAM Data */
	{
	    ADDRESS ptr = (vreg[18] << 8) || vreg[19];
	    if ( !((BYTE)++vreg[19]) )
		++vreg[18];
	    vdcRAM[ptr] = byte;
	}
	return;

      case 32:	/* Block Copy Start Address */
      case 33:
	return;
    }

    /*else
	reg = byte;*/
    return;
}


BYTE read_vdc(BYTE reg)
{
/*printf ("VDC read [%02X] %02x\n", reg, ((reg < 37) ? vreg[reg] : 0xff));*/

    if (reg == 31) {
	ADDRESS ptr = (vreg[18] << 8) || vreg[19];
	if ( !((BYTE)++vreg[19]) )
	    ++vreg[18];
	return (vdcRAM[ptr]);
    }
    return ((reg < 37) ? vreg[reg] : 0xff);
}


/* ------------------------------------------------------------------------- */

/*
 * Init & Control
 */


/* Xlib include files */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>

#include "x64.h"
/*#include "vmachine.h"*/
/*#include "extern.h"*/


/*
 * These were removed from the new system
 */

extern unsigned int width, height;
extern GC      cgcs[];
extern XColor  exact_defs[];
extern unsigned long colors[];
extern unsigned long bgcolors[];

/*
 *
 */

#ifndef __CEXTRACT__

extern void open_screen_win ( Window borderwin );
extern void vdc_set_vmode (  VDC *vdc, Window win, int mode );
extern void vdc_change_color ( int i, BYTE color );
extern void vdc_draw_screen ( ADDRESS addr );
static void draw_bitmap ( ADDRESS addr );
extern void vdc_scroll ( VDC *vdc, Window win, int xscroll, int yscroll, int oldx, int oldy );
extern void vdc_flush_queue ( void );

static void vdc_aargh ( VDC *vdc, Window win, ADDRESS address );
static void draw_nop ( VDC *vdc, Window win, ADDRESS address );
static void draw_text ( VDC *vdc, Window win, ADDRESS addr );
static void draw_chargen ( VDC *vdc, Window win, ADDRESS addr );
static void draw_bitmap_square ( VDC *vdc, Window win,  ADDRESS addr );

static void redraw_blank ( void );
static void redraw_text ( void );
static void redraw_bitmap ( void );

#endif /* __CEXTRACT__ */

/*
void    (*vdc_draw_square) ( VDC *vdc, Window win, ADDRESS );
void    (*vdc_draw_bitmap) ( VDC *vdc, Window win, ADDRESS );
void    (*vdc_redraw) ( void );
*/

extern Window  VDCborderwin;
static XImage *im, *im1, *im2;


#define QUEUESIZE 20000 /*TVR*/
static int     vdc_screenqueue[QUEUESIZE];
static int     vdc_queuesize;

static int     xoff, yoff;	/* scroll offsets */


/* ------------------------------------------------------------------------- */

/*
 * Initialize
 */

int  vdc_init(void)
{
    vdc.xpix = 640;
    vdc.ypix = 400;
    vdc.rows = 25;
    vdc.columns = 80;
    vdc.screen  = vdcRAM;
    vdc.color   = &vdcRAM[0x0800];
    vdc.chargen = &vdcRAM[0x2000];

    vdc_queuesize = 0;
    vdc_set_vmode(&vdc, VDCborderwin, V_TEXT);

    return (0);
}


void  vdc_set_vmode( VDC *vdc, Window win, int mode)
{
    int     i;

    if (debugflg)
	printf("VDC vmode set to %d\n", mode);

    switch (mode) {
      case V_TEXT:
	XMapWindow(display, win);
	vdc_draw_square = draw_text;
	vdc_draw_bitmap = draw_chargen;
	vdc_redraw = redraw_text;
	for (i = 0; i < MAX_COLORS; i++)
	    XSetBackground(display, cgcs[i], bgcolors[C_BG0]);
	break;

      case V_BITMAP:
	XMapWindow(display, win);
	vdc_draw_square = draw_bitmap_square;
	vdc_draw_bitmap = draw_bitmap;
	vdc_redraw = redraw_bitmap;
	break;

      case V_BLANK:
	XUnmapWindow(display, win);
	vdc_draw_square = vdc_aargh;
	vdc_draw_bitmap = vdc_aargh;
	vdc_redraw = redraw_blank;
	break;
    }
    redraw++;
}


/* ------------------------------------------------------------------------- */

/* interface functions to memory.c */

void    vdc_change_color(int i, BYTE color)
{
    XColor  exact_def;

    exact_def = exact_defs[color & 0xf];
    exact_def.pixel = bgcolors[i];

    XStoreColor(display, DefaultColormap(display, screen), &exact_def);
    return;

}


void    vdc_draw_screen(ADDRESS addr)
{
    vdc_screenqueue[vdc_queuesize++] = addr & 0x3fff;
}


void    vdc_scroll(VDC *vdc, Window win, int xscroll, int yscroll, int oldx, int oldy)
{
    int     xs = XPIX;
    int     ys = YPIX;

    if (xscroll & 8) {
	xoff = xscroll - 8;
	XClearArea(display, win, 0, 0, xoff, 0, False);
    } else {
	xoff = xscroll - 7;
	xs -= 0x10;
    }

    if (yscroll & 8) {
	yoff = yscroll - 11;
	if (yoff > 0)
	    XClearArea(display, win, 0, 0, 0, yoff, False);
    } else {
	yoff = yscroll - 7;
	ys -= 8;
    }
    if (verflg)
	printf("xs:%d ys:%d xoff:%d yoff:%d\n", xs, ys, xoff, yoff);

    XMoveResizeWindow(display, win, (width - xs) / 2, (height - ys) / 2, xs, ys);
    XMapWindow(display, win);
}


/* ------------------------------------------------------------------------- */

/*
 * Drawing
 */

/* draw chars in screen queue */

void    vdc_flush_queue(void)
{
    int     addr, i, j;

    /* printf ("queue %d\n", vdc_queuesize); */
    /* print chars in queue */

    for (i = 0; i < vdc_queuesize; i++) {
	addr = vdc_screenqueue[i];
	for (j = i + 1; j < vdc_queuesize; j++)
	    if (vdc_screenqueue[j] == addr)
		break;
	if (j == vdc_queuesize) {
	    (*vdc_draw_square)(&vdc, VDCborderwin, addr);
	}
    }
    vdc_queuesize = 0;
}

/* ------------------------------------------------------------------------- */

static void vdc_aargh(VDC *vdc, Window win, ADDRESS address)
{
    printf("AAAAARRRRGGHH!!!!! %x \n", (int) address);
}


/* draw functions called when square in screen have to be drawn */

static void    draw_nop(VDC *vdc, Window win, ADDRESS addr)
{
    return;
}


static void    draw_text(VDC *vdc, Window win, ADDRESS addr)
{
    int     x, y;

    x = addr % XSIZE;
    y = addr / XSIZE;
  /*  XPutImage(display, win, cgcs[vdc->color[addr] & 0xf],
	im, 0, scrbase[addr] << 3,
	xoff + (x << 3), yoff + (y << 3), 8, 8);*/
    printf ("DRAW %04x  row: %d col: %d  '%c'\n", addr, y, x, ' ');
}


static void    draw_bitmap_square(VDC *vdc, Window win, ADDRESS addr)
{
    int     x, y, color;
    addr &= 0x03ff;

    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    x = addr % XSIZE;
    y = addr / XSIZE;

    color = scrbase[addr] >> 4;
    XSetBackground(display, cgcs[color], colors[scrbase[addr] & 0xf]);
    XPutImage(display, win, cgcs[color], addr & 0x200 ? im2 : im1,
	0, (addr & 0x1ff) << 3, xoff + (x << 3), yoff + (y << 3), 8, 8);
}


/* redraws */

static void    redraw_blank(void)
{
    vdc_queuesize = 0;
    /* XClearArea(display, win, 0, 0, 0, 0, 0); */
}


static void    redraw_text(void)
{
    BYTE   *m = scrbase;
    BYTE   *c = vdc.color;
    int     x, y;

    vdc_queuesize = 0;

    im -> data = (char *) charbase;
    for (y = 0; y < YSIZE; y++)
	for (x = 0; x < XSIZE; x++) {
	    XPutImage(display, VDCborderwin, cgcs[*c++ & 0xf], im,
		0, *m++ * 8, xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


static void    redraw_bitmap(void)
{
    BYTE   *m = scrbase;
    int     x, y, color, i;

    vdc_queuesize = 0;
    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    for (i = y = 0; YSIZE > y; y++)
	for (x = 0; XSIZE > x; x++, i += 8) {
	    color = *m >> 4;
	    XSetBackground(display, cgcs[color], colors[*m++ & 0xf]);
	    XPutImage(display, VDCborderwin, cgcs[color], i & 0x1000 ? im2 : im1,
		0, i & 0xfff, xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


/*
 * bitmap draw functions
 */

static void    draw_chargen(VDC *vdc, Window win, ADDRESS addr)
{
    /* all chars, whose bitmap is modified must be redrawn */
    /* this does really slow the emulator down */

    int     i;
    addr = (addr >> 3) & 0xff;	/* changed char */
    for (i = 0; i < 2000; i++)
	if (scrbase[i] == addr)
	    vdc_screenqueue[vdc_queuesize++] = i;
}


static void    draw_bitmap(ADDRESS addr)
{
    vdc_screenqueue[vdc_queuesize++] = (addr >> 3) & 0x3fff;
}
