
/*
 * $Id: vic.c,v 1.10 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 *
 * $Log: vic.c,v $
 * Revision 1.10  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * VIC drawing routines renamed appropriately.
 * Recently introduced bug in Scroll-Redraw fixed.
 *
 * Revision 1.9  1994/12/12  16:59:44  jopi
 * New system
 *
 * Revision 1.8  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.7  1994/06/08  12:21:08  jopi
 * Patchlevel 2
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 * New debugger & arg parser and x128
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:39:47  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#include <stdio.h>

#include "macro.h"
#include "vmachine.h"
#include "extern.h"


/* global */

BYTE    vic[64];

BYTE   *scrbase = ram;		/* pointer to start of screen */
BYTE   *charbase = ram;		/* pointer to start of chargen */
BYTE   *bitmap1_base = ram;	/* pointer to upper half of bitmap */
BYTE   *bitmap2_base = ram;	/* pointer to lower half of bitmap */

BYTE    colorRam[0x400];

#ifdef C128
extern BYTE mmu[];
#endif

int     yscroll = 0;
int     xscroll = 0;

int     pow2[] = {
    0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80,
    0x100, 0x200, 0x400, 0x800, 0x1000, 0x2000, 0x4000, 0x8000
};


/* extern functions */

extern void    vic_set_vmode(int mode); /*temp */


/* hooks to draw char if needed */
static void (*screenwr_hook) ( ADDRESS, BYTE );

/* hooks to change draw display if chargen is changed */
static void (*matrix_hook1) ( ADDRESS, BYTE );
static void (*matrix_hook2) ( ADDRESS, BYTE );

/* local functions */

static void screenwr ( ADDRESS, BYTE );
static void chargenwr ( ADDRESS, BYTE );
static void matrixwr1 ( ADDRESS, BYTE );
static void matrixwr2 ( ADDRESS, BYTE );

BYTE   *calc_sprite_addr ( BYTE i );


/* local variables */

static int videoint;	/* interrupt register for vic */
static CLOCK rasteru;	/* time when next raster count reaches raster
			 * register */

int     vbank = 0;	/* video bank start address / 0x1000 (0, 4, 8 or 12) */
int     vmode = 0;	/* current video mode (see vmachine.h) */
int     charaddr;	/* chargen start address / 0x800 */
ADDRESS scraddr;	/* screen start address */

/* ------------------------------------------------------------------------- */

/* this is called in the middle of init_memmaps() */

void    set_video_hooks(void)
{
    int     bank;

    flush_queue();	/* Try to sync split pictures */

    switch (vmode) {
      case V_BLANK:
      case V_TILT:
	break;

      case V_TEXT:
      case V_ECTEXT:
      case V_MTEXT:
	/* check if screen is drawn from ram */
	bank = vbank + ((vic[0x18] >> 6) & 3);
	if ((bank & 7) != 1) {
	    screenwr_hook = memwr[bank];
	    memwr[bank] = screenwr;
	    if (verflg)
		printf("set screen hook to bank %d\n", bank);
	}
	/* chargen bank resolution is $800 (2k) */
	charaddr = (vbank << 1) + ((vic[0x18] >> 1) & 7);
	bank = charaddr >> 1;
	if ((bank & 7) != 1) {
	    matrix_hook1 = memwr[bank];
	    memwr[bank] = chargenwr;
	    if (verflg)
		printf("set chargen hook to bank %d chargen addr %x\n",
		    bank, charaddr * 0x800);
	}
	break;

      case V_BITMAP:
      case V_MBITMAP:
	/* check if screen is drawn from ram */
	bank = vbank + ((vic[0x18] >> 6) & 3);
	if ((bank & 7) != 1) {
	    screenwr_hook = memwr[bank];
	    memwr[bank] = screenwr;
	    if (verflg)
		printf("set screen hook to bank %d\n", bank);
	}
	bank = vbank + ((vic[0x18] >> 2) & 2);
	matrix_hook1 = memwr[bank];
	memwr[bank] = matrixwr1;
	if(verflg)
	    printf("set matrix1 hook to bank %d\n", bank); /*FCP*/

	/* chargen is visible $1000-$2000 and $9000-$a000 */
	bank |= 1;
	if ((bank & 7) != 1) {
	    matrix_hook2 = memwr[bank];
	    memwr[bank] = matrixwr2;
	    if(verflg)
		printf("set matrix2 hook to bank %d\n", bank); /*FCP*/
	}
	break;
    }
}


void    screen_init(void)
{
    ADDRESS addr;
    BYTE   *tmp1, *tmp2;
    int     tmp3, tmp4, tmp5;

    /* you are not supposed to understand this ... */

    /* first calculate screen start address */
    tmp1 = scrbase;
    scraddr = (vbank << 12) + ((vic[0x18] & 0xf0) << 6);
    if ((scraddr & 0x7000) != 0x1000) {
	scrbase = ram + scraddr;
    } else {
	scrbase = char_rom + (scraddr & 0x800);
    }

    /* then update pointers for chargen and bitmap */
    tmp2 = charbase;
    addr = (vbank << 12) + ((vic[0x18] & 0xe) << 10);
    if (debugflg)
	printf("charbase:%x\n", addr);
    bitmap1_base = ram + (addr & 0xe000);
    if ((addr & 0x7000) != 0x1000) {
	charbase = ram + addr;
	bitmap2_base = ram + ((addr & 0xf000) | 0x1000);
    } else {
	charbase = char_rom + (addr & 0x0800);
	bitmap2_base = char_rom;
    }

    /* set video mode */
    tmp3 = vmode;
    if (!(vic[0x11] & 0x10))
	vmode = V_BLANK;
    else
	switch (((vic[0x11] >> 5) & 3) | ((vic[0x16] >> 2) & 4)) {
	  case 0:
	    vmode = V_TEXT;
	    break;
	  case 1:
	    vmode = V_BITMAP;
	    break;
	  case 2:
	    vmode = V_ECTEXT;
	    break;
	  case 4:
	    vmode = V_MTEXT;
	    break;
	  case 5:
	    vmode = V_MBITMAP;
	    break;
	  case 3:
	  case 6:
	  case 7:	/* Illegal video mode ext. col & (multicolor |
			 * bitmap) */
	    vmode = V_TILT;
	    break;
	}

    tmp4 = yscroll;
    yscroll = vic[0x11] & 0xf;
    tmp5 = xscroll;
    xscroll = vic[0x16] & 0xf;

    if (scrbase != tmp1) {
	BYTE    t = vic[0x15];
	int     i = 0;

	while (t) {
	    if (t & 1)
		x_change_sprite(i, calc_sprite_addr(scrbase[0x3f8 + i]),
		    vic[0x1d] & pow2[i], vic[0x17] & pow2[i]);
	    t >>= 1;
	    i++;
	}
    }
    /* only if something is changed tell screen about it */
    if (charbase != tmp2 || scrbase != tmp1 || vmode != tmp3) {
	/* initialize memory & set video hooks */
	init_memmaps();
	vic_set_vmode(vmode);
    }
    if (tmp4 != yscroll || tmp5 != xscroll) {
	x_scroll(xscroll, yscroll, tmp5, tmp4);
    }
}

#define DRAW_SPRITE(i) 	x_move_sprite((i),\
		   vic[(i)<<1]+((vic[0x10]&pow2[i])?0x100:0),vic[((i)<<1)|1]);


/* ------------------------------------------------------------------------- */

void    store_vic(ADDRESS addr, BYTE byte)
{
    vic[addr] = byte;
    if (debugflg)
	printf("store vic[%x] %x\n", addr, (int) byte);

    switch (addr) {
      case 0:
      case 1:
      case 2:
      case 3:
      case 4:
      case 5:
      case 6:
      case 7:
      case 8:
      case 9:
      case 10:
      case 11:
      case 12:
      case 13:
      case 14:
      case 15:/* SPRITE COORDINATES */
	addr >>= 1;
	if (vic[0x15] & pow2[addr])
	    DRAW_SPRITE(addr);
	return;
      case 16:
	{
	    static  last_x_msb;
	    BYTE    t = last_x_msb ^ byte;
	    int     i = 0;
	    last_x_msb = byte;
	    while (t) {
		if (t & 1)
		    DRAW_SPRITE(i);
		t >>= 1;
		i++;
	    }
	    return;
	}

      case 17:/* $D011 */
	screen_init();
	update_raster(1);
	return;

      case 18:/* $D012 raster line compare */
	update_raster(1);
	return;

      case 19:
      case 20:/* Light Pen X,Y */
	return;

      case 21:/* $D015 SPRITE enable */
	{
	    static  last_spr_mask;
	    BYTE    t = last_spr_mask ^ byte;
	    int     i = 0;
	    last_spr_mask = byte;

	    while (t) {
		if (t & 1) {
		    if (byte & 1) {
			x_change_sprite
			    (i, calc_sprite_addr(scrbase[0x3f8 + i]),
			    vic[0x1d] & pow2[i],
			    vic[0x17] & pow2[i]);
			x_draw_sprite(i);
			DRAW_SPRITE(i);
		    } else
			x_undraw_sprite(i);
		}
		t >>= 1;
		byte >>= 1;
		i++;
	    }
	    return;
	}

      case 22:/* $D016 */
	screen_init();	/* possible multicolor mode */
	return;

      case 23:/* $D017 sprite expand y */
	{
	    static  last_exp_y;
	    BYTE    t = last_exp_y ^ byte;
	    int     i = 0;
	    last_exp_y = byte;

	    while (t) {
		if (t & 1)
		    x_change_sprite(i, calc_sprite_addr(scrbase[0x3f8 + i]),
			vic[0x1d] & pow2[i], vic[0x17] & pow2[i]);
		t >>= 1;
		i++;
	    }
	}
	return;

      case 24:/* $D018 video and char matrix base address */
	screen_init();
	return;

      case 25:/* $D019 IRQ flag register */
	return;
      case 26:/* $D01A IRQ mask register */
	/* light pen interrupt not implemented */
	if (byte & 8)
	    int_clk[I_LIGHT_PEN] = 0;
	else
	    int_clk[I_LIGHT_PEN] = 0;

	/* sprite to sprite collision not yet implemented */
	if (byte & 4)
	    int_clk[I_SS_COL] = 0;
	else
	    int_clk[I_SS_COL] = 0;

	/* sprite to background collision not yet implemented */
	if (byte & 2)
	    int_clk[I_SB_COL] = 0;
	else
	    int_clk[I_SB_COL] = 0;

	/* raster interrupts */
	if (byte & 1)
	    update_raster(1);
	else
	    int_clk[I_RASTER] = 0;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
	return;

      case 27:/* $D01B sprite to BG priority */
	return;
      case 28:/* $D01C sprite multicolor */
	return;
      case 29:/* $D01D sprite expand x */
	{
	    static  last_exp_x;
	    BYTE    t = last_exp_x ^ byte;
	    int     i = 0;
	    last_exp_x = byte;

	    while (t) {
		if (t & 1)
		    x_change_sprite(i, calc_sprite_addr(scrbase[0x3f8 + i]),
			vic[0x1d] & pow2[i], vic[0x17] & pow2[i]);
		t >>= 1;
		i++;
	    }
	}
	return;

      case 30:/* $D01E sprite collision */
	return;
      case 31:/* $D01F sprite to BG collision */
	return;

	/* colors */
      case 32:/* $D020 BORDER COLOR */
	x_change_color(C_BORDER, byte);
	return;
      case 33:/* $D021 BACKGROUND 0 */
	x_change_color(C_BG0, byte);
	return;
      case 34:/* $D022 BACKGROUND 1 */
	x_change_color(C_BG1, byte);
	return;
      case 35:/* $D023 BACKGROUND 2 */
	x_change_color(C_BG2, byte);
	return;
      case 36:/* $D024 BACKGROUND 3 */
	x_change_color(C_BG3, byte);
	return;
      case 37:/* sprite multicolor register 0 */
	x_change_color(C_SPRMUL0, byte);
	return;
      case 38:/* sprite multicolor register 1 */
	x_change_color(C_SPRMUL1, byte);
	return;
      case 39:/* sprite colours */
	x_change_color(C_SPR0, byte);
	return;
      case 40:
	x_change_color(C_SPR1, byte);
	return;
      case 41:
	x_change_color(C_SPR2, byte);
	return;
      case 42:
	x_change_color(C_SPR3, byte);
	return;
      case 43:
	x_change_color(C_SPR4, byte);
	return;
      case 44:
	x_change_color(C_SPR5, byte);
	return;
      case 45:
	x_change_color(C_SPR6, byte);
	return;
      case 46:/* $D02E */
	x_change_color(C_SPR7, byte);
	return;
#ifdef C128
      case 47:/* $D02F additional KBD rows */
	return;
      case 48:/* $D030 test, fast */
	return;
#endif
    }
}


BYTE    read_vic(ADDRESS addr)
{
    switch (addr) {
	case 17:	/* $D011 */
	return ((vic[addr] & 0x7f) | ((CURRENT_LINE(clk) & 0x100) >> 1));

      case 18:/* $D012 raster register */
	return (CURRENT_LINE(clk) & 0xff);

      case 19:
      case 20:/* Light Pen X,Y */
	return (0xff);

      case 25:/* $D019 interrupt register */
	{
	    BYTE    tmp;
	    if (clk > rasteru)
		update_raster(0);
	    tmp = videoint;
	    videoint = 0;
	    return (tmp);
	}

      case 26:/* $D01A interrupt control */
	return (vic[addr]);

      case 30:/* $D01E sprite sprite collision */
	return (vic[addr]);

      case 31:/* $D01F sprite background collision */
	return (vic[addr]);

#ifdef C128
      case 47:/* $D02F additional KBD rows */
	return (vic[47] | 0xf8);
#endif

      default:
	return ((addr < 0x31) ? vic[addr] : 0xff);  /* VIC IIe size */
    }
}


void store_colorram(ADDRESS addr, BYTE byte)
{
    if ((byte & 0xf) != (colorRam[addr] & 0xf)) {
        colorRam[addr] = byte;
        if (addr < 1000 &&
            (vmode == V_TEXT || vmode == V_ECTEXT || vmode == V_MTEXT))
            vic_draw_screen(addr);
    } /*else
        colorRam[addr] = byte;*/
}

BYTE read_colorram(ADDRESS addr)
{
  return colorRam[addr];
}


/* ------------------------------------------------------------------------- */

BYTE   *calc_sprite_addr(BYTE i)
{
    ADDRESS addr = (vbank << 12) + (i << 6);
    if ((addr & 0x7000) != 0x1000) {
	return (ram + addr);
    } else {
	return (char_rom + (addr & 0xfff));
    }
}


void    int_raster(void)
{
    if (debugflg)
	printf("raster interrupt\n");
    update_raster(1);
    videoint |= 0x80 | 1;
}


void    update_raster(int flag)
{
    int     lines = vic[0x12] + (((int) vic[0x11] & 0x80) << 1) - 
	(int)CURRENT_LINE(clk);
    if (lines < 0)
	lines += SCR_LINES;
    if (lines < 0)
	printf("strange raster %d %d %d %ld %ld curr line %ld\n",
	    lines, (int) vic[0x12], (int) vic[0x11], rasteru, clk,
	    CURRENT_LINE(clk));
    if (rasteru < clk)
	videoint |= 1;
    rasteru = clk + lines * CYCLES_PER_LINE;
    if (flag && (vic[0x1a] & 1)) {
	int_clk[I_RASTER] = rasteru;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
    }
}


static void screenwr(ADDRESS address, BYTE byte)
{
    /*
     * if scr is drawn from char_rom memwr is not hooked, so we can use
     * safely scraddr to determine if write hits screen area
     */

#ifdef C128
/*printf("%04X: screen %X%04X: %02x\n", PC, (mmu[0]>>6) & 3, address, byte);*/

    if ((mmu[0] ^ mmu[6]) & 0xC0) {	/* Check if not Video Bank */
/*printf("MISS!\n");*/
	(*screenwr_hook)(address, byte);
	return;
    }
#endif

    if ((address & 0xfc00) == scraddr) {
	/* do nothing unless ram is changed */
	if (byte != ram[address]) {
	    (*screenwr_hook)(address, byte);
	    if ((address & 0x3ff) < 1000)
		vic_draw_screen(address);
	    else
	     /* check for sprite origin */ if ((address & 0x3ff) > 0x3f7)
		x_change_sprite(address & 7, calc_sprite_addr(byte),
		    vic[0x1d] & pow2[address & 7],
		    vic[0x17] & pow2[address & 7]);
	}
    } else
	(*screenwr_hook)(address, byte);
}


static void chargenwr(ADDRESS address, BYTE byte)
{
    /* If chargen is changed on the fly, we must update screen. */
    (*matrix_hook1)(address, byte);
    if ((address >> 11) == charaddr)
	(*vic_draw_bitmap)(address);
}


static void matrixwr1(ADDRESS address, BYTE byte)
{
    /* These are used only in bitmap modes */
    (*matrix_hook1)(address, byte);
    (*vic_draw_bitmap)(address);
}


static void matrixwr2(ADDRESS address, BYTE byte)
{
    (*matrix_hook2)(address, byte);
    (*vic_draw_bitmap)(address);
}
