/*
 * $Id: vic.h,v 1.2 1994/12/12 16:59:44 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * Written by
 *   Hannu Helminen  (dm@stekt.oulu.fi)
 *
 *
 * $Log: vic.h,v $
 * Revision 1.2  1994/12/12  16:59:44  jopi
 * New system
 *
 * Revision 1.1  1994/08/10  17:46:41  jopi
 * Initial revision
 *
 *
 */

#ifndef X64_VIC_H
#define X64_VIC_H


#define MAX_COLORS 16

extern void store_vic(ADDRESS addr, BYTE byte);
extern void store_colorram(ADDRESS addr, BYTE byte);
extern BYTE read_colorram(ADDRESS addr);
extern BYTE read_vic(ADDRESS addr);
extern char *get_vic_bitmap();

#define PAL

#ifdef PAL

/* please check VicXExp etc */

/* expanded x size */
#define VicXSize (40*8)
#define VicYSize (25*8)

/* expaneded */
#define VicXExp 0x32
#define VicYExp 0x18

/* non-expanded */
#define VicXOff 0x36
#define VicYOff 0x1f

/* total size */
#define VicXTotal (63 * 8)
#define VicYTotal (200 + 112)

/* center, ie. which part should be in center of window */
#define VicXCenter (VicXExp + VicXSize/2)
#define VicYCenter (VicYExp + VicYSize/2)

#else
#error #define PAL
#endif

/*********************************************************************
* VIC register definitions
*********************************************************************/

enum D011 {
  RASCOMP = 0x80, EXTENDCOLOR = 0x40, BITMAP = 0x20, BLANK = 0x10,
  YEXPAND = 0x08, SMOOTHYMASK = 0x07
};
enum D016 {
  ALWAYS0 = 0x20, MULTICOLOR = 0x10, XEXPAND = 0x08, SMOOTHXMASK = 0x07
};
enum D017 {
  MATRIXMEMMASK = 0xf0, CHARMEMMASK = 0x0e
};
enum D019 /* also D01A */ {
  ANYIRQ = 0x80, LIGHTPENIRQ = 0x8, SPRITEIRQ = 0x04, BACKGROUNDIRQ = 0x02,
  RASTERIRQ = 0x01
};

struct VicReg {
  struct {BYTE x, y;} spritePos[8];
  BYTE spritePosMSB;
  BYTE rasterControl; /* D011 */
  BYTE rasterReg;     /* D012 */
  BYTE lightPenX;
  BYTE lightPenY;
  BYTE spriteEnable;
  BYTE vicControl;   /* D016 */
  BYTE spriteYExpand;/* D017 */
  BYTE memoryControl;/* D018 */
  BYTE irqFlag;      /* D019 */
  BYTE irqMask;
  BYTE spritePriority;
  BYTE spriteMultiColor;
  BYTE spriteXExpand;
  BYTE spriteCollision;
  BYTE backgroundCollision;
  BYTE borderColor;
  BYTE color[4];
  BYTE multicolor[2];
  BYTE spritecolor[8];
};

/*********************************************************************
* VIC structure
*********************************************************************/

typedef struct {
  int charCounter;
  int currx, curry;
  BYTE reg[0x40]; /* 0x2e first ones are used */
  BYTE bank;
  BYTE *colorRam;
  char *screen;
  int charCount;

  BYTE charBuf[0x28];
  BYTE colorBuf[0x28];
} VIC;

#define ASSERT(c) \
  if (!(c)) { \
    fprintf(stderr, "%s line %d: assertion failed\n", __FILE__, __LINE__); \
  }

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

#endif   /* X64_VIC_H */
