/*
 * $Id: vmachine.h,v 1.10 1994/12/12 16:59:44 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * This file contains useful stuff when you are creating
 * virtual machine like MCS6510 based microcomputer.
 *
 * Included are:
 *	o Interrupts
 *	o Colors
 *	o Video timing
 *
 *
 * Written by
 *   Vesa-Matti Puro (vmp@lut.fi)
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 *
 * $Log: vmachine.h,v $
 * Revision 1.10  1994/12/12  16:59:44  jopi
 * New system
 *
 * Revision 1.9  1994/08/10  18:34:13  jopi
 * More changeability
 *
 * Revision 1.8  1994/06/16  17:19:26  jopi
 *
 * Revision 1.7  1994/06/07  18:31:11  jopi
 * Patchlevel 2
 *
 * Revision 1.6  1994/02/18  16:10:57  jopi
 *
 * Revision 1.5  1994/01/26  16:08:37  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 *
 * Revision 1.3  93/06/21  13:39:50  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 *
 */

#ifndef X64_VMACHINE_H
#define X64_VMACHINE_H


#include "config.h"
#include "types.h"

/* INTERRUPS */

#define I_BRK		0	/* software interrupt */
/* VIC IRQ */
#define I_LIGHT_PEN	1
#define I_SS_COL	2	/* sprite to sprite collision */
#define I_SB_COL	3	/* sprite to background collision */
#define I_RASTER	4	/* raster compare */

/* CIA 1 IRQs */
#define I_CIA1FL	5	/* CIA 1 FLAG1 */
#define I_CIA1SP	6	/* CIA 1 serial port */
#define I_CIA1TOD	7	/* CIA 1 TOD alarm */
#define I_CIA1TA	8	/* CIA 1 timer A */
#define I_CIA1TB	9	/* CIA 1 timer B */

/* CIA 2 NMIs */
#define I_CIA2FL	10	/* CIA 2 FLAG 1 */
#define I_CIA2SP	11	/* CIA 2 serial port */
#define I_CIA2TOD	12	/* CIA 2 TOD alarm */
#define I_CIA2TA	13	/* CIA 2 timer A */
#define I_CIA2TB	14	/* CIA 2 timer B */

#define I_SPECX		15	/* special "interrupt" for x_loop */
#define I_RESTORE	16	/* restore key NMI */
#define I_RESET		17	/* reset in startup */
#define NUMOFINT	18


/*
 * Video Constants
 */

#define BORDER_WIDTH	4
#define MAX_COLORS	16

#define XSIZE		40
#define YSIZE		25
#define XPIX		320
#define YPIX		200

/* border, background & sprite colors */
#define C_BORDER	0
#define C_BG0		1
#define C_BG1		2
#define C_BG2		3
#define C_BG3		4
#define C_SPRMUL0	5
#define C_SPRMUL1	6
#define C_SPR0		7
#define C_SPR1		8
#define C_SPR2		9
#define C_SPR3		10
#define C_SPR4		11
#define C_SPR5		12
#define C_SPR6		13
#define C_SPR7		14

/* display modes */
#define V_BLANK		0
#define V_TEXT		1
#define V_MTEXT		2
#define V_ECTEXT	3
#define V_BITMAP	4
#define V_MBITMAP	5
#define V_TILT		6


/* Timing Constants
 *
 * Exact values are more complicated, but these approximations are
 *  enough for emulator's use
 */

#ifdef PAL
#define CYCLES_PER_SEC	985248
#define RFSH_PER_SEC	50
#define SCR_LINES	312
#define CYCLES_PER_LINE	63
#else

#define CYCLES_PER_SEC	1022730
#define RFSH_PER_SEC	60
#define SCR_LINES	262
#define CYCLES_PER_LINE	65
#endif

#define CYCLES_PER_RFSH (CYCLES_PER_SEC/RFSH_PER_SEC)
#define CURRENT_LINE(clk) (((clk)/CYCLES_PER_LINE) % SCR_LINES)


#endif  /* X64_VMACHINE_H */
