/*
 * $Id: x64.c,v 1.9 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta     (jopi@stekt.oulu.fi)
 *
 * Patch by
 *   Richard Hable (Richard.Hable@jk.uni-linz.ac.at) /RH/
 *
 *
 * $Log: x64.c,v $
 * Revision 1.9  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * More options for OpenWin routine.
 *
 * Revision 1.8  1994/12/12  16:59:44  jopi
 * New Vic emulation
 *
 * Revision 1.7  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.6  1994/02/18  16:06:03  jopi
 * *** empty log message ***
 *
 * Revision 1.5  1994/01/26  16:21:38  jopi
 * X64 version 0.2 PL 1
 * keyboard input latched
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * *** empty log message ***
 *
 * Revision 1.3  93/06/21  13:39:52  jopi
 *  X64 version 0.2 PL 0
 *
 */

#define XPIX 320
#define YPIX 200

#define BORDER_WIDTH 4
#define BITMAPDEPTH 1

#define MAXDOWN 16


/* Xlib include files */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <X11/cursorfont.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>

#include "cpu.h"	/* PC */
#include "vmachine.h"	/* VIC definitions */
#include "extern.h"
#include "x64.h"


#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif


/* Define Keyboard Buffer control */
#define ACK_UNREAD	1
#define ACK_DOWN	2
#define ACK_MODIFIER	4

#ifdef AUTOREPEAT
#define ACK_PRESSED	(ACK_UNREAD)
#else
#define ACK_PRESSED	(ACK_UNREAD | ACK_DOWN)
#endif


/* Bitmap data for icon */
#include "c64map.h"
#define icon_bitmap_bits dummy_bits
#define icon_bitmap_width dummy_width
#define icon_bitmap_height dummy_height


/*
 * functions
 */

static void get_colors ( void );
static Window OpenWindow ( int, int, int, int, char *, char * );
extern void open_screen_win ( Window borderwin );

extern void handle_mainwin_event ( XEvent *report );

#ifdef XDEBUGGER
extern void  debug_main ( void );
extern void  toggle_cntl_win ( void );
#endif


/*
 * Display and screen are used as arguments to nearly every Xlib routine
 * so it simplifies routine calls to declare them global; if there were
 * additional source files, these variables would be declared ``extern''
 * in them
 */

Display *display;
int      screen;
static Window  rootw;
Window   borderwin, VDCborderwin;
GC       gc;

char   *display_name = NULL;	/* X Server to connect to */

int      width, height;		/* for xscroll routine */


/* xdebugger */

/*int     halt, trace;*/
int     redraw;


/* ------------------------------------------------------------------------- */

/* keyboard */

BYTE    joy1, joy2;

typedef struct {
    KeySym  key;
    int     ack;
    int     time;
}  kbq;


kbq keyqueue[MAXDOWN];

static int     keys_down = 0;
static int     kbd_changed = 0;

keyconv keyconvmap[] = {
    {XK_Delete, 0, 1, 0},	/* del */
    {XK_Return, 0, 2, 0},
    {XK_Right, 0, 4, 0},
    {XK_F7, 0, 8, 0},
    {XK_F1, 0, 16, 0},
    {XK_F3, 0, 32, 0},
    {XK_F5, 0, 64, 0},
    {XK_Down, 0, 128, 0},

    {XK_3, 1, 1, 0},
    {XK_w, 1, 2, 0},
    {XK_a, 1, 4, 0},
    {XK_4, 1, 8, 0},
    {XK_z, 1, 16, 0},
    {XK_s, 1, 32, 0},
    {XK_e, 1, 64, 0},
    {XK_Shift_L, 1, 128, 0},

    {XK_5, 2, 1, 0},
    {XK_r, 2, 2, 0},
    {XK_d, 2, 4, 0},
    {XK_6, 2, 8, 0},
    {XK_c, 2, 16, 0},
    {XK_f, 2, 32, 0},
    {XK_t, 2, 64, 0},
    {XK_x, 2, 128, 0},

    {XK_7, 3, 1, 0},
    {XK_y, 3, 2, 0},
    {XK_g, 3, 4, 0},
    {XK_8, 3, 8, 0},
    {XK_b, 3, 16, 0},
    {XK_h, 3, 32, 0},
    {XK_u, 3, 64, 0},
    {XK_v, 3, 128, 0},

    {XK_9, 4, 1, 0},
    {XK_i, 4, 2, 0},
    {XK_j, 4, 4, 0},
    {XK_0, 4, 8, 0},
    {XK_m, 4, 16, 0},
    {XK_k, 4, 32, 0},
    {XK_o, 4, 64, 0},
    {XK_n, 4, 128, 0},

    {XK_plus, 5, 1, 0},
    {XK_p, 5, 2, 0},
    {XK_l, 5, 4, 0},
    {XK_minus, 5, 8, 0},
    {XK_period, 5, 16, 0},
    {XK_colon, 5, 32, 528},
    {XK_at, 5, 64, 528},	/* shifted @ gives 'grave' */
    {XK_comma, 5, 128, 0},

    {XK_sterling, 6, 1, 0},	/* pound sign */
    {XK_bar, 6, 1, 0},		/* pound sign */
    {XK_asterisk, 6, 2, 528},
    {XK_semicolon, 6, 4, 528},
    {XK_Home, 6, 8, 0},

#if 0
    {XK_Shift_R, 6, 16, 16},
#else
    {XK_Shift_R, 0, 0, 16},	/* invisible shift ;) */
#endif

    {XK_equal, 6, 32, 528},
#ifdef US_KBD
    { XK_asciicircum, 6,64,0 }, /* up arrow (US kbd) */
#else
    {XK_asciitilde, 6, 64, 0},	/* up arrow (scandinavian kbd) */
#endif
    {XK_slash, 6, 128, 528},

    {XK_1, 7, 1, 0},
    {XK_underscore, 7, 2, 0},	/* right arrow */
    {XK_Control_L, 7, 4, 0},	/* CNTRL */
    {XK_2, 7, 8, 0},
    {XK_space, 7, 16, 0},
    {XK_Alt_L, 7, 32, 0},	/* CBM */
    {XK_q, 7, 64, 0},
    {XK_Escape, 7, 128, 0},	/* STOP */

    /* some keys for DECstations */
    {XK_KP_F1, 0, 16, 0},	/* F1 */
    {XK_KP_F2, 0, 32, 0},	/* F3 */
    {XK_KP_F3, 0, 64, 0},	/* F5 */
    {XK_KP_F4, 0, 4, 0},	/* F7 */
    {XK_Find, 6, 8, 0},		/* Home */
    {XK_Insert, 0, 1, 16},	/* Inst */

#ifdef DECMAP
    {DXK_Remove, 0, 1, 0},	/* del (DECkeysym.h) */
    {XK_Help, 7, 128, 0},	/* because no real ESC-key on the keyboard /RH/ */
#endif

    /* shifted keys */
    {XK_BackSpace, 0, 1, 16},	/* Inst */
    {XK_Linefeed, 0, 2, 16},
    {XK_Left, 0, 4, 16},
    {XK_F8, 0, 8, 16},
    {XK_F2, 0, 16, 16},
    {XK_F4, 0, 32, 16},
    {XK_F6, 0, 64, 16},
    {XK_Up, 0, 128, 16},

    {XK_numbersign, 1, 1, 16},
    {XK_W, 1, 2, 16},
    {XK_A, 1, 4, 16},
    {XK_dollar, 1, 8, 16},
    {XK_Z, 1, 16, 16},
    {XK_S, 1, 32, 16},
    {XK_E, 1, 64, 16},
    /* { XK_Shift_L,	1,128,0 }, */

    {XK_percent, 2, 1, 16},
    {XK_R, 2, 2, 16},
    {XK_D, 2, 4, 16},
    {XK_ampersand, 2, 8, 16},
    {XK_C, 2, 16, 16},
    {XK_F, 2, 32, 16},
    {XK_T, 2, 64, 16},
    {XK_X, 2, 128, 16},

    {XK_quoteright, 3, 1, 16},	/* apostrophe */
    {XK_Y, 3, 2, 16},
    {XK_G, 3, 4, 16},
    {XK_parenleft, 3, 8, 16},
    {XK_B, 3, 16, 16},
    {XK_H, 3, 32, 16},
    {XK_U, 3, 64, 16},
    {XK_V, 3, 128, 16},

    {XK_parenright, 4, 1, 16},
    {XK_I, 4, 2, 16},
    {XK_J, 4, 4, 16},
    /* { XK_0,	4,8,16 }, */
    {XK_M, 4, 16, 16},
    {XK_K, 4, 32, 16},
    {XK_O, 4, 64, 16},
    {XK_N, 4, 128, 16},

    /* { XK_plus, 5,1,16 }, */
    {XK_P, 5, 2, 16},
    {XK_L, 5, 4, 16},
    /* { XK_minus, 5,8,16 }, */
    {XK_greater, 5, 16, 16},
    {XK_bracketleft, 5, 32, 16},
    {XK_grave, 5, 64, 16},	/* shifted @ */
    {XK_less, 5, 128, 16},

    /* { XK_sterling,	6,1,16 }, */
    {XK_backslash, 6, 1, 16},
    /* { XK_asterisk, 	6,2,16 }, */
    {XK_bracketright, 6, 4, 16},
    /* {XK_Home, 6,8,16 },
       {XK_Shift_R, 6,16,16 },
       {XK_equal, 6,32,16 }, */
#ifdef US_KBD
    { XK_asciitilde, 6,64,0 },		/* shift up arrow (US kbd) */
#else
    {XK_asciicircum, 6, 64, 0},		/* shift up arrow (scandinavian kbd) */
#endif
    {XK_question, 6, 128, 16},

    {XK_exclam, 7, 1, 16},
    /* { XK_underscore,	7,2,16 }, *//* right arrow */
    /* { XK_Control_L,	7,4,16 }, *//* CNTRL */
    {XK_quotedbl, 7, 8, 16},
    /* { XK_space, 7,16,16 },
       { XK_Alt_L, 7,32,16 }, *//* CBM */
    {XK_Q, 7, 64, 16},
    /* { XK_Escape, 7,128,16 }, *//* STOP */
    {0, 0, 0, 0}
};

#define CONV_KEYS (sizeof(keyconvmap)/sizeof(keyconv))


/* ------------------------------------------------------------------------- */

/* colors */
int has_colors;
unsigned long bgcolors[20];
/* colors for border, background & sprites */

XColor  exact_defs[MAX_COLORS];
unsigned long colors[MAX_COLORS];
GC      cgcs[MAX_COLORS];

static char   *cnames[] = {
    "Black",  "White",  "Red",  "Cyan",
    "Purple", "Green3", "Blue", "Yellow",
    "Orange", "Brown",  "Pink", "Dim Gray",
    "Gray", "Pale Green", "Light Blue", "Light Gray"
};

static char   *bwnames[] = {
    "Black", "White", "Dim Gray", "Light Gray",
    "Gray", "Gray", "Dim Gray", "Light Gray",
    "Gray", "Dim Gray", "Gray", "Dim Gray",
    "Gray", "Light Gray", "Gray", "Light Gray"
};


/* ------------------------------------------------------------------------- */

Bool   get_boolean (char *str) {
    return !strcmp (str, "true");
}


static void get_colors(void)
{
    int     def_depth = DefaultDepth(display, screen);
    XColor  exact_def, rgb_def;
    Colormap def_cmap = DefaultColormap(display, screen);
    int     i;
    XGCValues values;

    if (def_depth < 4) {
	printf("Cannot allocate colors.\n");
	exit(0);
    }
    /* allocate read/write colors for backgrounds & sprites */
    if (!XAllocColorCells(display, def_cmap, False, NULL, 0, bgcolors, 20)) {
	printf("Cannot allocate r/w colors. Please free the colormap.\n");
	exit(0);
    }
    /* allocate read only colors for characters in screen */
    for (i = 0; i < MAX_COLORS; i++) {
	if (!XAllocNamedColor
	    (display, def_cmap, cnames[i], &exact_def, &rgb_def)) {
	    printf("Cannot allocate color %s\n", cnames[i]);
	    exit(0);
	}
	values.foreground = exact_def.pixel;

	cgcs[i] = XCreateGC(display, rootw, GCForeground, &values);

	colors[i] = exact_def.pixel;
	exact_defs[i] = exact_def;
	exact_defs[i].flags = DoRed | DoGreen | DoBlue;
    }

    values.foreground = BlackPixel(display, screen);
    values.background = WhitePixel(display, screen);
}


/*
 * General purpose window opening routine.
 */

static Window OpenWindow(win_x, win_y, width, height, win_name, icon_name)
     int win_x, win_y;
     int width, height;
     char *win_name, *icon_name;
{
    Window        win;
    XTextProperty windowName, iconName;
    XWMHints      wm_hints;
    XClassHint    class_hints;
    XSizeHints    size_hints;		/* Preferred sizes for window man */
    XSetWindowAttributes xswattr;

    unsigned int  display_width, display_height;
    int  flags   = PSize | PMinSize;
    int  screen;


    /* Display size is a member of display structure */
    screen = DefaultScreen(display);
    display_width = DisplayWidth(display, screen);
    display_height = DisplayHeight(display, screen);


    if (height < 0 || width < 0) {
	width =  XPIX + 60;
	height = YPIX + 60;
    }
    if (win_x >= 0 && win_y >= 0)	/* no random placement */
	flags |= PPosition;

    if (win_x == 0 && win_y == 0) {	/* center */
	win_x = (display_width  - width) / 2;
	win_y = (display_height - height) / 2;
    }

/*
    xswattr.background_pixel = WhitePixel(display, screen);
    xswattr.border_pixel = BlackPixel(display, screen);
    xswattr.backing_store = WhenMapped;
*/

    xswattr.background_pixel = bgcolors[C_BORDER];
    xswattr.cursor        = XCreateFontCursor(display, XC_gumby);
    xswattr.backing_store = Always;
    xswattr.event_mask    = ExposureMask | KeyReleaseMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask |
	FocusChangeMask | EnterWindowMask | LeaveWindowMask;

    win = XCreateWindow(display, RootWindow(display, screen), win_x, win_y,
	      width, height, 0,
	      CopyFromParent, CopyFromParent, CopyFromParent,
	      /*CWBackingStore |*/ CWBackPixel | CWEventMask | CWCursor,
	      &xswattr);

    if (!win) {
	fprintf(stderr, "could not open window called %s", win_name);
	x_exit();
	exit(0);
    }

    size_hints.flags = flags;
    size_hints.width = width;
    size_hints.height = height;
    size_hints.min_width =  XPIX;
    size_hints.min_height = YPIX;
    size_hints.x = win_x;
    size_hints.y = win_y;

    XSetStandardProperties(display, win, win_name, win_name,
	None, NULL, 0, &size_hints);

    /* */
    if (XStringListToTextProperty(&win_name, 1, &windowName) == 0) {
	(void) fprintf
	    (stderr, "structure allocation for windowName failed.\n");
	exit(-1);
    }
    if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
	(void) fprintf
	    (stderr, "structure allocation for iconName failed.\n");
	exit(-1);
    }
    /* */

    class_hints.res_name = "x64";
    class_hints.res_class = "C64";
    XSetClassHint(display, win, &class_hints);

    /*XSelectInput(display, win, ExposureMask | KeyPressMask |
		 ButtonPressMask | StructureNotifyMask);*/

    /* */
    wm_hints.initial_state = NormalState;
    wm_hints.input = True;
    wm_hints.flags = StateHint | InputHint;

    XSetWMProperties(display, win, &windowName, &iconName, NULL, 0,
	&size_hints, &wm_hints, &class_hints);
    /* */


    /* Display window */
    XMapWindow(display, win);
    return(win);
}


/* ------------------------------------------------------------------------- */

void    x_start(void)
{
    /*Pixmap icon_pixmap;*/
    int winw, winh;

#ifdef XDEBUGGER
    debug_main ();

#else
   /* Connect to X server */
    if ((display = XOpenDisplay(display_name)) == NULL) {
	(void) fprintf (stderr, "cannot connect to X server %s\n",
	     XDisplayName(display_name));
	exit(-1);
    }
#endif

    /* Get screen size from display structure macro */
    screen = DefaultScreen(display);
    rootw = RootWindow(display, screen);

    get_colors();

    width = winw = XPIX + 60;
    height= winh = YPIX + 60;

#ifdef C128
#ifdef USE_VDC
    VDCborderwin = OpenWindow(-1, -1, 640, 400, "VDC@ C128 Emulator", "C128 80c");
    printf("*** The blank screen is NOT error. ***\n\n");
    printf("Great! As you seem to be interested,\n could you write the rest of VDC stuff ?\n\n");
#endif
    borderwin = OpenWindow(-1, -1, winw, winh, "VIC@ C128 Emulator", "C128 40c");
#else
    borderwin = OpenWindow(0, 0, winw, winh, "C64 Emulator", "C64");
#endif
    open_screen_win (borderwin);
}	       /* End x_start */


/* ------------------------------------------------------------------------- */


void    x_exit(void)
{
    int     i;

    XAutoRepeatOn(display);
    for (i = 0; i < MAX_COLORS; i++)
	XFreeGC(display, cgcs[i]);
    XCloseDisplay(display);
}

/* ------------------------------------------------------------------------- */


#ifndef XDEBUGGER

void    x_loop(void)
{
    XEvent  report;	/* Structure for event information */

    if (redraw) {
	vic_redraw();
	redraw = 0;
    }

#ifndef SETOWN
	flush_queue();
#endif

    while (XEventsQueued(display, QueuedAfterFlush) > 0 
	   /* || (halt && !trace) */ ) {
	XNextEvent(display, &report);

	/* Main Window Event */
	if (report.xany.window == borderwin) {
	    handle_mainwin_event(&report);
	    continue;
	}
    }	/* End while */

    /*if (trace) {
	int_clk[I_SPECX] = clk;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif
    }*/
}

#endif


/* ------------------------------------------------------------------------- */


static void handle_debugwin_event(XEvent *report)
{
    static char buffer[20];
    KeySym  key;
    XComposeStatus compose;
    int     count;
    int     show_keyarr();

    /* printf("debugwin event\n"); */
    switch (report -> type) {
      case KeyPress:
	count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	switch (key) {
	  case XK_Up:
	    joy1 |= 1;
	    break;
	  case XK_Down:
	    joy1 |= 2;
	    break;
	  case XK_Left:
	    joy1 |= 4;
	    break;
	  case XK_Right:
	    joy1 |= 8;
	    break;
	  case XK_Return:
	    joy1 |= 16;
	    break;
	}
	if (debugflg)
	    printf("press joy1 %x\n", joy1);
	break;

      case KeyRelease:
	count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

	switch (key) {
	  case XK_Up:
	    joy1 &= ~1;
	    break;
	  case XK_Down:
	    joy1 &= ~2;
	    break;
	  case XK_Left:
	    joy1 &= ~4;
	    break;
	  case XK_Right:
	    joy1 &= ~8;
	    break;
	  case XK_Return:
	    joy1 &= ~16;
	    break;
	}
	if (debugflg)
	    printf("release joy1 %x\n", joy1);
	break;
    }
}


void    handle_mainwin_event(XEvent *report)
{
    switch (report -> type) {

      case Expose:
	if (debugflg)
	    printf("Expose %d\n", report -> xexpose.count);

	/* Unless this is the last contiquous expose, don't draw the window */
	if (report -> xexpose.count != 0)
	    break;

	/* Place text in window */
	XClearWindow(display, borderwin);
	(*vic_redraw)();
	break;

      case ConfigureNotify:
	if (debugflg)
	    printf("Config notify\n");
	/* window has been resized; change width and height */
	height = report -> xconfigure.height;
	width = report -> xconfigure.width;
	break;

      case ButtonPress:
	switch (report -> xbutton.button) {
	  case Button1:
	    /* debugflg = !debugflg; */
	    break;
	  case Button2:
#ifndef AUTOREPEAT
	    XAutoRepeatOn(display);
#endif
	    /* -- drop to monitor; continue with x or C-d */
	    mon(PC);
#ifndef AUTOREPEAT
	    XAutoRepeatOff(display);
#endif
	    break;
#ifdef XDEBUGGER
	  case Button3:
	    /* open debugging window */
	    toggle_cntl_win();
#endif
	}
	break;

      case KeyPress:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count, i;

	    count = XLookupString(&report -> xkey, buffer, 20, &key, &compose);

#ifdef KBDBUG
	    printf("\ncount %d Key %d %x Compose %x\n",
		   count, (int) *buffer, key, compose);
#endif
	    if (key == XK_Tab)	/* Restore key (Esc-Tab = break) */
		restore();

	    /*
	     * On a multitasking system input has to be buffered since
	     * we don't have direct contact to the keyboard.
	     * First search for last occurrence of current key to work out
	     * possible autorepeat.
	     */
	    for (i = keys_down; i && (key != keyqueue[--i].key););

#ifdef KBDBUG
	    printf("%2d old key %02x + %ld\n",
		   i, keyqueue[i].key & 0xffff,
		   report -> xkey.time - keyqueue[i].time);
#endif
	    if (key == keyqueue[i].key &&
		(report -> xkey.time - keyqueue[i].time < REPEAT_DELAY)) {
		keyqueue[i].time = report -> xkey.time;
		keyqueue[i].ack  = ACK_PRESSED;
	    }
	    else if (keys_down < MAXDOWN) {
		    if (keyqueue[i].ack)
			i = keys_down;

		    keyqueue[i].key = key;
		    keyqueue[i].time = report -> xkey.time;
		    keyqueue[i].ack  = ACK_PRESSED;
#ifdef AUTOREPEAT
		    /* mark modifiers (non-repeating) */ 
		    if (key == XK_Shift_L || key == XK_Shift_R ||
			key == XK_Control_L || key == XK_Return)
			keyqueue[i].ack |= (ACK_MODIFIER | ACK_DOWN);
#endif /* repeat */

		    kbd_changed++;
		    keys_down++;
	    }

#ifdef KBDBUG
	    printf("KEY FOUND [%d/%d] Key %02x  - %ld\n",
		   i, keys_down, key & 0xffff, report -> xkey.time);

	    for (i = 0; i < MAXDOWN; i++)
		printf("%02x ", keyqueue[i].key & 0xffff);
	    printf("\n");
#endif


	    /* this hack puts keystrings directly into char buffer */
#ifdef KBD_HACK
#ifdef KBD_HACKALL
	      if( ( key == XK_Escape ) /*RH*/
		|| ( key == XK_Help )
		|| ( key == XK_Home )
		|| ( key == XK_Control_L )
		|| ( key == XK_Alt_L )
		|| ( key == XK_Delete )
		|| ( key == XK_Right )
	 	|| ( key == XK_Down )
		|| ( key == XK_Find )
		|| ( key == XK_Insert )
		|| ( (key >= XK_F1) && (key <= XK_F8) )  /* F1..F8 */
		|| ( key == XK_Left )
		|| ( key == XK_Up ) ) {;} else {
#endif
	     if (count>0) { if( ((*buffer & 0xc0) == 0x40)&&((*buffer
	       &0x1f) <0x1e) ) *buffer ^= 0x20; ram[0x0277 + ram[0xc6]++] =
	       *buffer; }

#ifdef KBD_HACKALL
	 } /*RH*/
#endif
#endif

#ifdef KBDBUG2
	printf("queue %d \n", keys_down);
	for (i = 0; i < MAXDOWN; i++)
	    printf("%02x%c  ", keyqueue[i].key, keyqueue[i].ack + '@');
	printf("\n");

	printf("%2d keys: %02X %02X %02X %02X  %02X %02X %02X %02X\n",
	       keys_down, keyarr[0], keyarr[1], keyarr[2], keyarr[3],
	       keyarr[4], keyarr[5], keyarr[6], keyarr[7]);
#endif

	    break;
	}


      case KeyRelease:
	{
	    static char buffer[20];
	    KeySym  key;
	    XComposeStatus compose;
	    int     count, i;

	    count = XLookupString((XKeyEvent *) report, buffer,
				  20, &key, &compose);

#ifdef AUTOREPEAT
	    /* With autorepeat, process only modifiers released. */ 
	    if (key != XK_Shift_L && key != XK_Shift_R &&
		key != XK_Control_L && key != XK_Return)
		break;
#endif /* repeat */

#ifdef KBDBUG
	    printf("release %d Key %d %x\n", count, (int) *buffer, key);
#endif
	    /*
	     * search for (last occurrence of) current key
	     * and set a flag it's released
	     */
	    for (i = keys_down; i >= 0; i--)
		if (key == keyqueue[i].key) {
		    keyqueue[i].ack &= ~ACK_DOWN;
		    ++kbd_changed;
#ifndef KBDBUG
		    if (debugflg)
#endif
			printf("KEY UP    [%d/%d] %ld\n",
			       i, keys_down, report -> xkey.time);
		    /*break;*/
		}

	    break;
	}


#ifndef AUTOREPEAT
      case FocusIn:
      case EnterNotify:

	XAutoRepeatOff(display);
	break;
#endif /* repeat */

      case FocusOut:
      case LeaveNotify:
	{
	    register int i;

	    /*
	     * Entering to the screen subwindow triggers this too, thus,
	     * it must be filtered out.
	     */

	    if (!report -> xcrossing.subwindow) {
		XAutoRepeatOn(display);
#ifdef KBDBUG
		printf("no focus\n");
#endif

	    /* this is just to recover from server's panic
	     * (may be obsolete already)
	     */

		for (i = 0; i < MAXDOWN; i++) {
		    keyqueue[i].key = 0;
		    keyqueue[i].ack = 0;
		}

		for (i = 0; i < 8; i++)
		    keyarr[i] = 0;
		keys_down = 0;
	    }
	}
	break;

      case ClientMessage:
	flush_queue();
	break;

      default:
	/*
	 * All events selected by StructureNotifyMask except ConfigureNotify
	 * are thrown away here, since nothing is done with them
	 */
	break;
    }	       /* End switch */
}

/* ------------------------------------------------------------------------- */

int     x_get_keyarr(void)
{
    KeySym  key;
    int     i,j;

    /* This routine is called once for each IRQ. */


/*    if (kbd_changed || (update >0 && !--update)) {
	kbd_changed = 0;
*/

	/* clean up entries that have became obsolete */

	for (i = j = 0; i < keys_down; i++) {
	    if (keyqueue[j].ack & (ACK_DOWN | ACK_UNREAD))
		j++;
	    if (i > j) {
		if (keyqueue[j].ack & (ACK_DOWN | ACK_UNREAD))
		{   keyqueue[j].key = keyqueue[i].key;
		    keyqueue[j].ack = keyqueue[i].ack;
		    keyqueue[j].time= keyqueue[i].time;
#ifdef KBDBUG
		    printf("Que %d -> %d\n", i, j);
#endif
		}
		else {
		    keyqueue[i].key = 0;
		    keyqueue[i].ack = 0;
#ifdef KBDBUG
		    printf("Remove [%d/%d] %02x\n", i, keys_down, keyqueue[i].key);
#endif
		}
	    }
	} /* for */

#if 0
    if (j)
	update = 10;		/* re-check after delay */
#endif

    keys_down = j;


#ifdef KBDBUG
	printf("queue (-%d) %d \n", i-j, keys_down);
	for (i = 0; i < MAXDOWN; i++)
	    printf("%02x%c  ", keyqueue[i].key, keyqueue[i].ack + '@');
	printf("\n");
#endif

	/*
	 * re-build keyboard matrix
	 */

	for (i = 0; i < 8; i++)
	    keyarr[i] = 0;


	/* read keypress event buffer */

	for (j = 0; j < keys_down && keyqueue[j].ack; j++) {
	    key = keyqueue[j].key;
	    keyqueue[j].ack &=  ~ACK_UNREAD;

	    for (i = 0; i < CONV_KEYS; i++) {
		if (key == keyconvmap[i].sym) {
		    keyarr[keyconvmap[i].row] |= keyconvmap[i].keybit;
		    if (keyconvmap[i].shift == 16)
			keyarr[6] |= 16;  /* add right shift */
		    else
			keyarr[6] &= ~16;
#ifdef KBDBUG
		    printf("Keyarr: %02X\n", key);
#endif
		    break;
		}
	    }
	}

#ifdef KBDBUG
	/*if (debugflg)
	  show_keyarr();*/

	printf("%2d keys: %02X %02X %02X %02X  %02X %02X %02X %02X\n",
	       keys_down, keyarr[0], keyarr[1], keyarr[2], keyarr[3],
	       keyarr[4], keyarr[5], keyarr[6], keyarr[7]);
#endif


/*    }*/ /* changed */

  return (kbd_changed);
}
