/*
 * $Id: xdebug.c,v 1.4 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 * X11-based 65xx tracer/debugger interface.
 * Also the tree of X resources is built here.
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *
 *
 * $Log: xdebug.c,v $
 * Revision 1.4  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * Added new options: -notraps -brevision.
 * Debugger status display added.
 *
 * Revision 1.3  1994/12/12  16:59:44  jopi
 * Added display of Memory Address, Memory Data, and Clock Cycles.
 *
 * Revision 1.2  1994/06/16  17:19:26  jopi
 * Xrm replaced with XtResources
 *
 * Revision 1.1  1994/02/18  18:26:14  jopi
 * Initial revision
 *
 *
 */


#ifdef __hpux
#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#endif


#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>     /* Get standard string definitions. */

/*
 * And then the same again by hand...
 * If you are unfortunate enough not to have XPointer etc defined by your
 * system, you have to define by hand all types that are missing:
 */

#ifdef X_NOT_STDC_ENV
typedef char * XPointer;
/*typedef void *XtPointer;*/
#define XtOffsetOf(s_type,field) XtOffset(s_type*,field)
#endif

#ifndef XtNjumpProc
#define XtNjumpProc "jumpProc"
#endif


/*** String definitions. ***/
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Xaw/Cardinals.h>

/*** Simple Widgets ***/
#include <X11/Xaw/Grip.h>
#include <X11/Xaw/List.h>
/* #include <X11/Xaw/Panner.h> */
/* #include <X11/Xaw/Repeater.h> */
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/StripChart.h>
#include <X11/Xaw/Toggle.h>

/*** Menus ***/
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

/*** Text Widgets ***/
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Text.h>
#include <X11/Xaw/Paned.h>

#include "cpu.h"
#include "macro.h"
#include "vmachine.h"	/* I_SPECX */
#include "extern.h"
#include "misc.h"
#include "x64.h"
 

/*
 * Variables
 */

XtAppContext    app_con;
extern Display *display;
extern Window   borderwin;

static int      halt, trace;

static ADDRESS  current_point;
static char     buf[256];



enum Register {
    R_AC, R_XR, R_YR, R_SP, R_PC, R_SR, R_EFF, R_DL, R_CLK, NO_REGISTERS
};

static char *registerLabels[] = {
    "AC:", "XR:", "YR:", "SP:", "PC:", "SR:", "EA>", "DL>", "Clk"
};


enum DState { DS_Run, DS_Halt, DS_Trace, DS_Step, DS_Break, DS_Mon };
enum DState  state;

static char *StatusLabels[] = {
    "Running.", "Halted.", "Trace Mode.", "Execute.", "Break.", "Monitor Break."
};



/*
 * Functions
 */

extern void  end64 ( void );
extern void  debug_main ( void );
extern void  toggle_cntl_win ( void );

static void  Syntax ( XtAppContext, char * );
void   handle_mainwin_event ( XEvent * );

static void  ThumbProc ( Widget bar, XtPointer client, XtPointer ptr );
static void  ScrollProc ( Widget bar, XtPointer client, XtPointer ptr );
static void  MenuBar ( Widget parent );
static void  RegisterBar ( Widget parent );

static void UpdateRegisters ( void );
static void UpdateStatus ( enum DState );
static void UpdateButtons ( void );

static void  set_asm ( ADDRESS );
static void  draw_asm ( ADDRESS );


/*
 * Widgets
 */

Widget registers[NO_REGISTERS];
Widget stopButton, traceButton, nextButton;
Widget view;
Widget ASMbar;
Widget status1;
Widget toplevel;


/*
 * All application resources must reside in a structure. The structure
 * contains space where the application resources will be stored.
 */

AppResources app_resources;

/*
 * The resources specification contains:
 *
 * The name and class of the resource.
 * The representation type of the resource.
 * The location in the structure to keep the value of this resource.
 * The default value for this resource.
 *
 * This information is passed into XtGetApplicationResources().
 */

static XtResource resources[] = {
 {"basicName",	"BasicName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, basicName),	XtRString, (XPointer)BASICNAME},
 {"kernelName",	"KernelName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, kernelName),	XtRString, (XPointer)KERNELNAME},
 {"charName",	"CharName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, charName),	XtRString, (XPointer)CHARGENNAME},
 {"ramName",	"RamName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, ramName),	XtRString, (XPointer)RAMNAME},
 {"exromName",	"ExromName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, exromName),	XtRString, (XPointer)EXROMNAME},
 {"module",	"Module",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, module),	XtRString, (XPointer)"module"},
 {"directory",	"Directory",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, directory),	XtRString, (XPointer)PROJECTDIR},
 {"reuName",	"ReuName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, reuName),	XtRString, (XPointer)0},
 {"kernelRev",	"KernelRev",	XtRString,    sizeof(XtRString),
      XtOffsetOf (AppResources, kernelRev),	XtRString, (XPointer)0},
 {"basicRev",	"BasicRev",	XtRString,    sizeof(XtRString),
      XtOffsetOf (AppResources, basicRev),	XtRString, (XPointer)0},
 {"colors",	"Colors",	XtRBoolean, sizeof(Boolean),
      XtOffsetOf (AppResources, colors),	XtRImmediate, (XtPointer)True},
 {"asm", 	"Asm",		XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, asmflag),	XtRImmediate, (XtPointer)False},
 {"hex", 	"Hex",		XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, hexflag),	XtRImmediate, (XtPointer)True},
 {"debug", 	"Debug",	XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, debugflag),	XtRImmediate, (XtPointer)False},
 {"verbose",	"Verbose",	XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, verboseflag),	XtRImmediate,(XtPointer)False},
 {"test", 	"Test",		XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, testflag),	XtRImmediate, (XtPointer)False},
 {"notraps", 	"NoTraps",	XtRBoolean, sizeof (XtRBoolean),
      XtOffsetOf (AppResources, notraps),	XtRImmediate, (XtPointer)False},

 {"programName","ProgramName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, programName),	XtRString, (XPointer)0},
 {"floppyName",	"FloppyName",	XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, floppyName),	XtRString, (XPointer)0},
 {"printCommand","PrintCommand",XtRString, sizeof(XtRString),
      XtOffsetOf (AppResources, printCommand),	XtRString, (XPointer)"lp"},
};


/*
 * Options
 */

static XrmOptionDescRec options[] = {
    {"-fn", 		"*font", 	XrmoptionSepArg, 	(XPointer)0},
    {"-display",	".display", 	XrmoptionSepArg, 	(XPointer)0},
    {"-name",		"*name", 	XrmoptionSepArg, 	(XPointer)0},
    {"+colors",		"*colors", 	XrmoptionNoArg, 	(XPointer)"false"},
    {"-colors",		"*colors", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-basicName",	"basicName",	XrmoptionSepArg, 	(XPointer)0},
    {"-kernelName",	"kernelName",	XrmoptionSepArg, 	(XPointer)0},
    {"-ramName",	"ramName",	XrmoptionSepArg, 	(XPointer)0},
    {"-exromName",	"exromName",	XrmoptionSepArg, 	(XPointer)0},
    {"-module",		"*module", 	XrmoptionSepArg, 	(XPointer)0},
    {"-directory",	"*directory",	XrmoptionSepArg, 	(XPointer)0},
    {"-load",		"*programName",	XrmoptionSepArg, 	(XPointer)0},
#ifdef REU
    {"-reuName",	"*reuName", 	XrmoptionSepArg, 	(XPointer)0},
#endif
#ifdef PATCH_ROM
    {"-revision",	"*kernelRev",	XrmoptionSepArg, 	(XPointer)0},
    {"-brevision",	"*basicRev",	XrmoptionSepArg, 	(XPointer)0},
#endif
    {"-asm",	 	"*asm", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-hex",	 	"*hex", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"+hex",	 	"*hex", 	XrmoptionNoArg, 	(XPointer)"false"},
    {"-debug",	 	"*debug", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-verbose",	"*verbose", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-test",	 	"*test", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-traps",		"*notraps", 	XrmoptionNoArg, 	(XPointer)"false"},
    {"-notraps",	"*notraps", 	XrmoptionNoArg, 	(XPointer)"true"},
    {"-floppy",		"*floppyName", 	XrmoptionSepArg, 	(XPointer)0},
    {"-printCommand",	"*printCommand",XrmoptionSepArg,	(XPointer)0},
};


/*
 * Xdebugger Resources
 */

String fallback_resources[] = {
    "*fileButton.label:      File",
    "*optionButton.label:    Options",

    "*stop.Label:            Stop",	/* Cont - Stop - Stop  */
    "*trace.Label:           Trace",	/* Step - Run  - Trace */
    "*next.Label:            Break",	/* Next - Next - Break */
    "*stop.width:            135",
    "*trace.width:           135",
    "*next.width:            135",
    "*view.allowVert:        True",
    "*view.forceBars:        True",
    "*view.width:            200",
    "*view.height:           320", 
    "*hbox.orientation:      Horizontal",
    "*hbox.borderWidth:      0",
    "*hbox.min:              28",
    "*hbox.max:              28",
    "*vbox.orientation:      Vertical",
    "*vbox.borderWidth:      0",   /* Registers and buttons */
    "*smeBSB*borderWidth:    1",
    "*Label*borderWidth:     0",   /* Register name boxes */
    "*bar.orientation:       Vertical",
    "*bar.length:            320",
    "*bar.shown:             0.02",

    "*foreground:	     blue",
    "*background:	     light blue",
    "*Label.foreground       black",
    NULL,
};


/* ------------------------------------------------------------------------- */

void    toggle_cntl_win(void)
{
    static int      debugstate = 0;

    if ((debugstate = !debugstate)) {
	XtRealizeWidget (toplevel);
	UpdateRegisters();
	set_asm(PC);
    } else {
	XtUnrealizeWidget (toplevel);
    }
}


void    x_loop(void)
{
    XEvent  report;	/* Structure for event information */


    if (halt || trace) {
	UpdateRegisters();
	set_asm(PC);
    }

    if (redraw) {
	(*vic_redraw)();
	redraw = 0;
    }

#ifndef SETOWN
    flush_queue();
#endif

    while ((halt && !trace) || XtAppPending (app_con)) {
	XtAppNextEvent(app_con, &report);

	/* Main Window Event */
	if (report.xany.window == borderwin) {
	    handle_mainwin_event(&report);
	    continue;
	}
	/* Xdebugger Event */
	XtDispatchEvent (&report);

	state = (halt ? DS_Halt : trace ? DS_Trace : DS_Run);
	UpdateStatus(state);

    }	/* End while */

    if (trace) {
	int_clk[I_SPECX] = clk;
	find_next_int();
#ifdef NEW_TIMER_CODE
	update_intr_alarm();
#endif

	if (halt)      /* single step excution */
	    trace = 0;

	state = (halt ? DS_Halt : trace ? DS_Trace : DS_Run);
	UpdateStatus(state);
    }
}


static void UpdateRegisters (void)
{
    int eff;

    sprintf (buf, "$%02X", AC);
    XtVaSetValues (registers [R_AC],
		   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", XR);
    XtVaSetValues (registers [R_XR],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", YR);
    XtVaSetValues (registers [R_YR],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", SP);
    XtVaSetValues (registers [R_SP],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%04X", PC);
    XtVaSetValues (registers [R_PC],
                   XtNstring, buf, 
		   NULL);

    sprintf (buf, "$%02X", GET_SR());
    XtVaSetValues (registers [R_SR],
                   XtNstring, buf, 
		   NULL);

    /* Memory */

    eff = eff_address(PC, 1);
    if (eff >= 0)
	sprintf (buf, "$%04X", eff);
    else
	sprintf (buf, "----");

    XtVaSetValues (registers [R_EFF],
                   XtNstring, buf, 
		   NULL);

    if (eff >= 0)
	sprintf (buf, "$%02X", LOAD(eff));
    else
	sprintf (buf, "--");

    XtVaSetValues (registers [R_DL],
		   XtNstring, buf, 
		   NULL);

    /* Clock */

    sprintf (buf, "%ld", clk);
    XtVaSetValues (registers [R_CLK],
		   XtNstring, buf, 
		   NULL);
}


/* Debugger Status */

static void UpdateStatus (enum DState dst)
{
    static int s = -1;
    if (s == (int)dst)
	return;

    XtVaSetValues (status1,
		   XtNlabel, StatusLabels[dst],
		   NULL);
    s = dst;
}


static void UpdateButtons (void)
{
    /*
     * Box widget rezises buttons by default
     */
    XtVaSetValues (stopButton,
                   XtNlabel, halt ? "Continue" : "Stop" , 
		   XtNwidth, 135,
		   NULL);

    XtVaSetValues (traceButton,
                   XtNlabel, halt ? "Step" : trace ? "Run" : "Trace" , 
		   XtNwidth, 135,
		   NULL);
/*
    XtVaSetValues (nextButton,
                   XtNlabel, (halt || trace) ? "Next" : "Break" , 
		   XtNwidth, 135,
		   NULL);
*/
}


static void QuitCb(Widget w, XtPointer clientData, XtPointer callData)
{
    end64 ();
}


/* ------------------------------------------------------------------------- */

static void StopCb(Widget w, XtPointer clientData, XtPointer callData)
{
    /*Edit self = (Edit)clientData;
    CrEditCall call = (CrEditCall)callData;
    AppSelectUnset(self->app);
    EditInsert(self,call->rect.x,call->rect.y,call->z);*/

    halt = !halt;

    UpdateButtons ();
    UpdateRegisters();
    set_asm(PC);
}


static void TraceCb(Widget w, XtPointer clientData, XtPointer callData)
{
    /*Edit self = (Edit)clientData;
    CrEditCall call = (CrEditCall)callData;
    AppSelectUnset(self->app);
    EditInsert(self,call->rect.x,call->rect.y,call->z);*/

#if 0
    if (halt) {
	trace = 1;	/* Step */
    }
    else {
	trace = !trace;
    }
#else
	trace = !trace;
#endif

    UpdateButtons ();
    UpdateRegisters();
    set_asm(PC);
}


static void NextCb(Widget w,XtPointer clientData,XtPointer callData)
{
    /*Edit self = (Edit)clientData;
    CrEditCall call = (CrEditCall)callData;
    AppSelectUnset(self->app);
    EditInsert(self,call->rect.x,call->rect.y,call->z);*/

   /*
    * Halt, Trace: if "Next" is pressed when JSR is to be excuted, tracing
    * is turned off until exution returns back, or the entry is removed from
    * stack. This button has no use in Run mode.
    */

    if (halt) {
	/* Next */
	trace = 1;		/* Step */
    }
    else {
	if (trace) {
	    /*trace = 0;*/
	}
	else {
	    printf("Run/Stop-Restore.\n");
	    restore();
	}
    }

    UpdateRegisters();
    set_asm(PC);
}



/* ------------------------------------------------------------------------- */

/*
 * Assembly Window
 */

static void ThumbProc (Widget bar, XtPointer client, XtPointer ptr)
{
    float percent = *(float*) ptr;
    ADDRESS p = ((int) (percent * 0xffff)) & 0xffff; 
    /*printf ("thumb %f\n", percent);*/
    draw_asm (p);
    return;
}


static void ScrollProc (Widget bar, XtPointer client, XtPointer ptr)
{
    int position = (int) ptr;

    printf ("position %d\n", position);
    if (position < 0) {
	draw_asm (current_point + 0x10);
    } else {
	draw_asm (current_point - 0x10);
    }
    return;
}


void set_asm (/*Widget bar,*/ ADDRESS p)
{
    float percent = (p & 0xffff) / 65535.0;
    XawScrollbarSetThumb (ASMbar, percent, 0.02);
    draw_asm (p);
}


void draw_asm (ADDRESS p)
{
    Widget  sink;
    char    buf2[8192];
    short   lines, height;  /* NOT ints */
    int i;


    buf2[0] = '\0';
    current_point = p;

    XtVaGetValues (view, 
	 XtNheight, &height, 
	 XtNtextSink, &sink,
	 NULL);
    lines = XawTextSinkMaxLines (sink, height);


    for (i = 0; i <= lines; i++) {
	p &= 0xffff;
	sprintf (buf, "%04X %s  %-s\n", p, 
		 sprint_ophex (p), sprint_opcode(p, 1));
	strcat (buf2, buf);
	p += clength[lookup[LOAD(p)].addr_mode];
    }

    XtVaSetValues (view,
                   XtNstring, buf2,
		   NULL);
}


/* ------------------------------------------------------------------------- */

static void MenuBar (Widget parent)
{
    Widget hbox, fileMenu, optionMenu, shell, use;

    hbox = XtVaCreateManagedWidget 
	("hbox", boxWidgetClass, parent, 
	 NULL);
    
    fileMenu = XtVaCreateManagedWidget 
	("fileButton", menuButtonWidgetClass, hbox,
	 XtNmenuName,"fileMenu",
	 NULL);

    shell = XtVaCreatePopupShell 
	("fileMenu", simpleMenuWidgetClass, fileMenu, 
	 NULL);

    use = XtVaCreateManagedWidget
        ("load",smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback(use,XtNcallback,AttributesCb,(XtPointer)self);*/

    use = XtVaCreateManagedWidget
	("quit",smeBSBObjectClass, shell,
         NULL);
    XtAddCallback (use, XtNcallback, QuitCb, (XtPointer)0);


    optionMenu = XtVaCreateManagedWidget 
	("optionButton", menuButtonWidgetClass, hbox,
	 XtNmenuName,"optionMenu",
	 NULL);

    shell = XtVaCreatePopupShell 
	("optionMenu", simpleMenuWidgetClass, optionMenu, 
	 NULL);

    use = XtVaCreateManagedWidget
        ("attributes",smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback(use,XtNcallback,AttributesCb,(XtPointer)self);*/

    XtVaCreateManagedWidget
        ("refresh",smeBSBObjectClass, shell,
         NULL);
    /*XtAddCallback (refresh,XtNcallback,RefreshCb,(XtPointer)self);*/
}


static void RegisterBar (Widget parent)
{
    int i;
    Widget form;
    Widget label, lastlabel;

    form = parent 
	/*XtVaCreateManagedWidget("form", formWidgetClass, parent, NULL)*/;

    /*
     * Registers
     */

    for (lastlabel = form, i = 0; i < NO_REGISTERS; i++, lastlabel = label) {
	label = XtVaCreateManagedWidget (
	     "register", labelWidgetClass, form, 
	     XtNlabel, registerLabels[i], 
	     XtNfromVert, lastlabel,
	     NULL);
	registers[i] = XtVaCreateManagedWidget (
	     "value",asciiTextWidgetClass, form,
	     XtNtype,XawAsciiString,
	     XtNeditType,XawtextEdit,
	     XtNfromHoriz, label,
	     XtNfromVert, lastlabel,
	     NULL);
    }

    /*
     * Debugger Status
     */

    status1 = XtVaCreateManagedWidget (
	"status", labelWidgetClass, form, 
	XtNlabel, StatusLabels[DS_Run],
	XtNfromVert, lastlabel,
	NULL);

    /*
     * Buttons
     */

    stopButton = XtVaCreateManagedWidget (
	"stop", commandWidgetClass, form,
	 XtNfromVert, status1,
	 NULL);
    XtAddCallback (stopButton, XtNcallback, StopCb, NULL);

    traceButton = XtVaCreateManagedWidget (
	"trace", commandWidgetClass, form,
	 XtNfromVert, stopButton,
	 NULL);
    XtAddCallback(traceButton, XtNcallback, TraceCb, NULL);
/*
    nextButton = XtVaCreateManagedWidget (
	"next", commandWidgetClass, form,
	XtNfromVert, traceButton,
	NULL);
    XtAddCallback(nextButton, XtNcallback, NextCb, NULL);
*/
}


/* ------------------------------------------------------------------------- */

/*
 *  Parse arguments and X resources
 */

void   arg_parser(int argc, char **argv)
{
    toplevel = XtAppInitialize(&app_con, "X64", options, XtNumber(options),
			       &argc, argv, fallback_resources, NULL, ZERO);

    if (argc != 1)
	Syntax(app_con, argv[0]);

    XtGetApplicationResources(toplevel, (XtPointer) &app_resources,
			      resources, XtNumber(resources), NULL, ZERO );
}


void   debug_main(void)
{
    Widget form, pane;
    Widget vbox;


    /*
     * Vertical Pane of widgets 
     */
    pane = XtCreateManagedWidget("pane", panedWidgetClass, toplevel, NULL, 0);

    /*
     * Menu bar
     */
    MenuBar (pane);

    form = XtVaCreateManagedWidget("form", formWidgetClass, pane, NULL);

    /*
     * Assembly listing
     */
#if 0
    view = XtVaCreateManagedWidget
	("view", viewportWidgetClass,form, 
	 XtNfromVert, hbox,
	 NULL);
#endif
    ASMbar = XtVaCreateManagedWidget("bar", scrollbarWidgetClass,form, NULL);
    XtAddCallback(ASMbar, XtNjumpProc, ThumbProc, NULL);
    XtAddCallback(ASMbar, XtNscrollProc, ScrollProc, NULL);

    view = XtVaCreateManagedWidget
	("view", asciiTextWidgetClass, form, 
	 XtNfromHoriz, ASMbar,
	 /*XtNstring, "hmm\nplh\n",*/
	 NULL);

    vbox = XtVaCreateManagedWidget 
	("vbox", boxWidgetClass, form, 
	 XtNfromHoriz, view,
	 NULL);
    
    RegisterBar (vbox);

    /*XtRealizeWidget(toplevel);*/

    display = XtDisplay (toplevel);
    /*XtAppMainLoop(app_con);*/
}


static void  Syntax (XtAppContext appc, char *call)
{
    int i;

    XtDestroyApplicationContext(appc);

    fprintf(stderr, "\nUsage: %s\n", call);
    for (i = 0; i < XtNumber(options); i++) {
	if (options[i].argKind == XrmoptionNoArg) {
	    fprintf(stderr, "    [ %s ]\n", options[i].option);
	} else {
	    fprintf(stderr, "    [ %s <value>]\n", options[i].option);
	}
    }
    fprintf(stderr, "\n    All options may be abbreviated.\n\n");
    exit(1);
}
