/*
 * $Id: xscreen.c,v 1.8 1995/04/01 07:54:09 jopi Exp $
 *
 * This file is part of Commodore 64 emulator.
 * See README for copyright notice
 *
 *
 * Written by
 *   Jarkko Sonninen (sonninen@lut.fi)
 *   Jouko Valta (jopi@stekt.oulu.fi)
 *
 * $Log: xscreen.c,v $
 * Revision 1.8  1995/04/01  07:54:09  jopi
 * X64 0.3 PL 0
 * VIC drawing routines renamed appropriately.
 *
 * Revision 1.7  1994/12/12  16:59:44  jopi
 * New Vic Emulation
 *
 * Revision 1.6  1994/06/16  17:19:26  jopi
 * Code reorganized and cleaned up
 *
 * Revision 1.5  1994/01/26  16:26:56  jopi
 * X64 version 0.2 PL 1
 *
 * Revision 1.4  1993/11/10  01:55:34  jopi
 * reu, asm and disk directory fixed
 * REL_ADDR macro and 1541 made more portable
 *
 * Revision 1.3  93/06/21  13:40:02  jopi
 *  X64 version 0.2 PL 0
 *
 * Revision 1.2  1993/06/13  08:21:50  sonninen
 * *** empty log message ***
 *
 * 
 */


#include "config.h"

/* Xlib include files */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>
#include <sys/ioctl.h>

#include "macro.h"
#include "vmachine.h"
#include "extern.h"
#include "x64.h"


/*
 * These are removed from the new system
 */

extern unsigned int width, height;
extern BYTE    colorRam[];
extern GC      cgcs[];
extern XColor  exact_defs[];
extern unsigned long colors[];
extern unsigned long bgcolors[];


/*
 *
 */

#ifndef __CEXTRACT__

extern void open_screen_win ( Window borderwin );
extern void vic_set_vmode ( int mode );

static void draw_nop ( void );
static void draw_text ( ADDRESS addr );
static void draw_mtext ( ADDRESS addr );
static void draw_ectext ( ADDRESS addr );
static void draw_chargen ( ADDRESS addr );

static void draw_bitmap ( ADDRESS addr );
static void draw_bitmap_square ( ADDRESS addr );
static void draw_mbitmap_square ( ADDRESS addr );

static void redraw_blank ( void );
static void redraw_tilt ( void );
static void redraw_text ( void );
static void redraw_mtext ( void );
static void redraw_ectext ( void );
static void redraw_bitmap ( void );
static void redraw_mbitmap ( void );

#endif /* __CEXTRACT__ */


void    (*vic_draw_square) ( ADDRESS );
void    (*vic_draw_bitmap) ( ADDRESS );
void    (*vic_redraw) ( void );


static Window  win;
static XGCValues xgcv;
static XImage *im, *im1, *im2;

static GC      mgc;

/* sprites */
static Window  sprite_win[8];
static GC      spritegc[8];
static int     sprexp[256];
static BYTE    sprexpl[256], sprexph[256];
static XImage *sprim, *sprim2;
static GC      shapeGC;
static Pixmap  sprshape;
static int shape;	/* shape extension available ? */

#define QUEUESIZE 20000 /*TVR*/
static int     screenqueue[QUEUESIZE];
static int     queuesize;

static int     xoff, yoff;	/* scroll offsets */
static BYTE    mc_mask[5][256];	/* multicolor masks */


/* ------------------------------------------------------------------------- */

void open_screen_win(Window borderwin)
{
    int     i, j, k, t;
    XGCValues values;

    XSetWindowAttributes wattr;
    int     shape_event_base, shape_error_base;

    wattr.backing_store = Always;
    wattr.background_pixel = bgcolors[C_BORDER];
    wattr.event_mask = ExposureMask | KeyReleaseMask | KeyPressMask |
	ButtonPressMask | StructureNotifyMask;

    wattr.backing_store = Always;
    wattr.bit_gravity = CenterGravity;
    wattr.win_gravity = CenterGravity;
    wattr.background_pixel = bgcolors[C_BG0];
    win = XCreateWindow(display, borderwin, 30, 30, XPIX, YPIX, 0,
	CopyFromParent, CopyFromParent, CopyFromParent,
	CWWinGravity | CWBitGravity | CWBackingStore | CWBackPixel,
	&wattr);
    /* XMapWindow (display, win); */


    sprim = XCreateImage(display, NULL, 1, XYBitmap, 0,
	(char *) ram, 24, 42, 8, 0);
    sprim -> bitmap_bit_order = MSBFirst;
    sprim -> byte_order = MSBFirst;

    sprim2 = XCreateImage(display, NULL, 1, XYBitmap, 0,
	(char *) ram, 48, 42, 8, 0);
    sprim2 -> bitmap_bit_order = MSBFirst;
    sprim2 -> byte_order = MSBFirst;

    sprshape = XCreatePixmap(display, win, 48, 42, 1);
    shapeGC = XCreateGC(display, sprshape, 0, &xgcv);
    XSetBackground(display, shapeGC, 0);

    for (i = 0; i < 8; i++) {
	values.foreground = bgcolors[C_SPR0 + i];
	spritegc[i] = XCreateGC(display, borderwin, GCForeground, &values);
    }

    mgc = XCreateGC(display, borderwin, 0, &values);

    wattr.backing_store = Always;
    wattr.save_under = True;

    /* open window for each sprite */
    for (i = 0; i < 8; i++) {
	sprite_win[i] =
	    XCreateWindow(display, win, 0, 0, 48, 42, 0,
	    CopyFromParent, CopyFromParent, CopyFromParent,
	    CWBackingStore | CWSaveUnder, &wattr);
	XFillRectangle(display, sprite_win[i], spritegc[i], 0, 0, 48, 42);
	/* XMapWindow (display, sprite_win[i]); */
    }

    /* image for chargen (font) */
    im = XCreateImage(display, NULL, 1, XYBitmap, 0,
	(char *) charbase, 8, 2048, 8, 0);
    im -> bitmap_bit_order = MSBFirst;
    im -> byte_order = MSBFirst;

    /* images for bitmap mode */
    im1 = XCreateImage(display, NULL, 1, XYBitmap, 0,
	(char *) charbase, 8, 4096, 8, 0);
    im1 -> bitmap_bit_order = MSBFirst;
    im1 -> byte_order = MSBFirst;

    im2 = XCreateImage(display, NULL, 1, XYBitmap, 0,
	(char *) charbase, 8, 4096, 8, 0);
    im2 -> bitmap_bit_order = MSBFirst;
    im2 -> byte_order = MSBFirst;

#ifdef SHAPE
    if (!XShapeQueryExtension(display, &shape_event_base, &shape_error_base)) {
	shape = False;
	printf("Screen win: Shape not ok\n");
    }
#endif

    sprexp[0] = 0;
    for (i = 1; i <= 0x80; i <<= 1)
	for (j = 0; j < i; j++) {
	    sprexp[j | i] = sprexp[j] | i * i | i * i * 2;
	    sprexpl[j | i] = LOWER (sprexp[j | i]);
	    sprexph[j | i] = UPPER (sprexp[j | i]);
	}


    for (j = 0; j < 0x100; j++) {
	for (t = j, k = 0; k < 4; k++) {
	    for (i = 0; i < 4; i++)
		mc_mask[i][j] >>= 2;
	    mc_mask[t & 3][j] |= 0xC0;
	    t >>= 2;
	}
	/*printf ("%d :  %3d %3d %3d %3d\n", j, 
	  mc_mask[0][j], mc_mask[1][j], mc_mask[2][j], mc_mask[3][j]);*/
    }

    vic_set_vmode(V_BLANK);

}


/* ------------------------------------------------------------------------- */

/* interface functions to memory.c */

void    x_change_color(int i, BYTE color)
{
    XColor  exact_def;

    exact_def = exact_defs[color & 0xf];
    exact_def.pixel = bgcolors[i];

    XStoreColor(display, DefaultColormap(display, screen), &exact_def);
    return;

}


void    vic_draw_screen(ADDRESS addr)
{
    screenqueue[queuesize++] = addr & 0x3ff;
}


static void    draw_bitmap(ADDRESS addr)
{
    screenqueue[queuesize++] = (addr >> 3) & 0x3ff;
}


void    vic_set_vmode(int mode)
{
    int     i;

    if (debugflg)
	printf("vmode set to %d\n", mode);

    flush_queue();	/* Try to sync split pictures */

    switch (mode) {
      case V_TEXT:
	XMapWindow(display, win);
	vic_draw_square = draw_text;
	vic_draw_bitmap = draw_chargen;
	vic_redraw = redraw_text;
	for (i = 0; i < MAX_COLORS; i++)
	    XSetBackground(display, cgcs[i], bgcolors[C_BG0]);
	break;

      case V_MTEXT:
	XMapWindow(display, win);
	vic_draw_square = draw_mtext;
	vic_draw_bitmap = draw_chargen;
	vic_redraw = redraw_mtext;
	break;

      case V_ECTEXT:
	XMapWindow(display, win);
	vic_draw_square = draw_ectext;
	vic_draw_bitmap = draw_chargen;
	vic_redraw = redraw_ectext;
	break;

      case V_BITMAP:
	XMapWindow(display, win);
	vic_draw_square = draw_bitmap_square;
	vic_draw_bitmap = draw_bitmap;
	vic_redraw = redraw_bitmap;
	break;

      case V_MBITMAP:
	XMapWindow(display, win);
	vic_draw_square = draw_mbitmap_square;
	vic_draw_bitmap = draw_bitmap;
	vic_redraw = redraw_mbitmap;
	break;

      case V_BLANK:
	XUnmapWindow(display, win);
	vic_draw_square = aargh;
	vic_draw_bitmap = aargh;
	vic_redraw = redraw_blank;
	break;

      case V_TILT:
	XResizeWindow(display, win, XPIX, YPIX);
	XMapWindow(display, win);
	vic_draw_square = aargh;
	vic_draw_bitmap = aargh;
	vic_redraw = redraw_tilt;
	break;
    }
    redraw++;
}


void    x_scroll(int xscroll, int yscroll, int oldx, int oldy)
{
    int     xs = XPIX;
    int     ys = YPIX;

    if (xscroll & 8) {
	xoff = xscroll - 8;
	XClearArea(display, win, 0, 0, xoff, 0, False);
    } else {
	xoff = xscroll - 7;
	xs -= 0x10;
    }

    if (yscroll & 8) {
	yoff = yscroll - 11;
	if (yoff > 0)
	    XClearArea(display, win, 0, 0, 0, yoff, False);
    } else {
	yoff = yscroll - 7;
	ys -= 8;
    }
    if (verflg)
	printf("xs:%d ys:%d xoff:%d yoff:%d\n", xs, ys, xoff, yoff);

    XMoveResizeWindow(display, win, (width - xs) / 2, (height - ys) / 2, xs, ys);
    XMapWindow(display, win);
}


void    x_move_sprite(int i, int x, int y)
{
    XMoveWindow(display, sprite_win[i & 7], x - 24, y - 50);
}


void    x_draw_sprite(int i)
{
    XMapWindow(display, sprite_win[i & 7]);
}


void    x_undraw_sprite(int i)
{
    XUnmapWindow(display, sprite_win[i & 7]);
}


void    x_change_sprite(int i, BYTE * q, int expx, int expy)
{
    int     xs;
    int     ys;
    int     j;
    BYTE    buf[256], *s, *r;

    if (expx) {
	if (expy) {
	    xs = 48;
	    ys = 42;
	    for (s = q, r = buf, j = 0; j < 21; j++, r += 12) {
		r[6] = r[0] = sprexph[*s];
		r[7] = r[1] = sprexpl[*s++];
		r[8] = r[2] = sprexph[*s];
		r[9] = r[3] = sprexpl[*s++];
		r[10] = r[4] = sprexph[*s];
		r[11] = r[5] = sprexpl[*s++];
	    }
	} else {
	    for (s = q, r = buf, j = 0; j < 42; j++) {
		*r++ = sprexph[*s];
		*r++ = sprexpl[*s++];
	    }
	    xs = 48;
	    ys = 21;
	}

	XResizeWindow(display, sprite_win[i], xs, ys);

	sprim2 -> data = (char *) buf;

#ifdef SHAPE
	if (shape) {
	    XSetForeground(display, shapeGC, 0);
	    XFillRectangle(display, sprshape, shapeGC, 0, 0, xs, ys);
	    XSetForeground(display, shapeGC, 1);
	    XPutImage(display, sprshape, shapeGC, sprim2, 0, 0, 0, 0, xs, ys);
	    XShapeCombineMask(display, sprite_win[i], ShapeBounding,
		0, 0, sprshape, ShapeSet);
	    XFillRectangle(display, sprite_win[i], spritegc[i], 0, 0, xs, ys);
	} else
#endif

	    XPutImage(display, sprite_win[i], spritegc[i], sprim2,
		0, 0, 0, 0, xs, ys);
    } else {
	if (expy) {
	    xs = 24;
	    ys = 42;
	    for (s = q, r = buf, j = 0; j < 21; j++, r += 6) {
		r[3] = r[0] = *s++;
		r[4] = r[1] = *s++;
		r[5] = r[2] = *s++;
	    }
	    sprim -> data = (char *) buf;
	} else {
	    xs = 24;
	    ys = 21;
	    sprim -> data = (char *) q;
	}

	XResizeWindow(display, sprite_win[i], xs, ys);

#if SHAPE
	if (shape) {
	    XSetForeground(display, shapeGC, 0);
	    XFillRectangle(display, sprshape, shapeGC, 0, 0, xs, ys);
	    XSetForeground(display, shapeGC, 1);
	    XPutImage(display, sprshape, shapeGC, sprim, 0, 0, 0, 0, xs, ys);
	    XShapeCombineMask(display, sprite_win[i], ShapeBounding,
		0, 0, sprshape, ShapeSet);
	    XFillRectangle(display, sprite_win[i], spritegc[i], 0, 0, xs, ys);

	} else
#endif

	    XPutImage(display, sprite_win[i], spritegc[i], sprim,
		0, 0, 0, 0, xs, ys);
    }
}


/*---------------------------------------------------------------------------*/

/* draw chars in screen queue */

void    flush_queue(void)
{
    int     addr, i, j;

    /* printf ("queue %d\n", queuesize); */
    /* print chars in queue */

    for (i = 0; i < queuesize; i++) {
	addr = screenqueue[i];
	for (j = i + 1; j < queuesize; j++)
	    if (screenqueue[j] == addr)
		break;
	if (j == queuesize) {
	    (*vic_draw_square)(addr);
	}
    }
    queuesize = 0;
}

/* ------------------------------------------------------------------------- */

/*
 * draw functions called when square in screen have to be drawn
 */

void    aargh(ADDRESS address)
{
    printf("AAAAARRRRGGHH!!!!! %x \n", (int) address);
}


static void    draw_nop(void)
{
    return;
}


static void    draw_text(ADDRESS addr)
{
    int     x, y;

    x = addr % XSIZE;
    y = addr / XSIZE;
    XPutImage(display, win, cgcs[colorRam[addr] & 0xf],
	im, 0, scrbase[addr] << 3,
	xoff + (x << 3), yoff + (y << 3), 8, 8);
    /*printf ("DRAW %04x  row: %d col: %d  '%c'\n", addr, y, x, im);*/
}


static void    draw_mtext(ADDRESS addr)
{
    int     x, y, i, j, k, t;
    BYTE    m, buf[8], *p;

    x = addr % XSIZE;
    y = addr / XSIZE;

    x = xoff + (x << 3);
    y = yoff + (y << 3);
    addr &= 0x03ff;
    m = scrbase[addr];

    im -> data = (char *)buf;

    XSetFunction (display, mgc, GXcopy);
    XSetForeground(display, mgc, bgcolors[C_BG0]);
    for (p =  charbase + (m << 3), j = 0; j < 8; j++)
	buf[j] = mc_mask[0][*p++];
    XPutImage(display, win, mgc, im, 0, 0, x, y, 8, 8);

    XSetFunction (display, mgc, GXor);
    XSetForeground(display, mgc, bgcolors[C_BG1]);
    for (p =  charbase + (m << 3), j = 0; j < 8; j++)
	buf[j] = mc_mask[1][*p++];
    XPutImage(display, win, mgc, im, 0, 0, x, y, 8, 8);

    XSetForeground(display, mgc, bgcolors[C_BG2]);
    for (p =  charbase + (m << 3), j = 0; j < 8; j++)
	buf[j] = mc_mask[2][*p++];
    XPutImage(display, win, mgc, im, 0, 0, x, y, 8, 8);

    XSetForeground(display, mgc, exact_defs[colorRam[addr] & 0x7].pixel);
    for (p =  charbase + (m << 3), j = 0; j < 8; j++)
	buf[j] = mc_mask[3][*p++];
    XPutImage(display, win, mgc, im, 0, 0, x, y, 8, 8);

#if 0
    for (k = 0; k < 4; k++) {
	switch (k) {
	  case 0:
	    XSetForeground(display, mgc, bgcolors[C_BG0]);
	    break;
	  case 1:
	    XSetForeground(display, mgc, bgcolors[C_BG1]);
	    break;
	  case 2:
	    XSetForeground(display, mgc, bgcolors[C_BG2]);
	    break;
	  case 3:
	    XSetForeground(display, mgc,
		exact_defs[colorRam[addr] & 0x7].pixel);
	    break;
	}
      
	for (p =  charbase + (m << 3), j = 0; j < 8; j++) {
	    buf[j] = mc_mask[k][*p++];
	    t = charbase[(m << 3) + j];
	    for (i = 6; i >= 0; i -= 2) {
		if ((t & 3) == k) {
		    XDrawPoint(display, win, mgc, x + i, y + j);
		    XDrawPoint(display, win, mgc, x + i + 1, y + j);
		}
		t >>= 2;
	    }
	}
    }
#endif
}


static void    draw_ectext(ADDRESS addr)
{
    int     x, y, color;
    BYTE    m;

    addr &= 0x3ff;

    x = addr % XSIZE;
    y = addr / XSIZE;

    im -> data = (char *) charbase;
    color = colorRam[addr] & 0xf;
    m = scrbase[addr];
    switch (m >> 6) {
      case 0:
	XSetBackground(display, cgcs[color], bgcolors[C_BG0]);
	break;
      case 1:
	XSetBackground(display, cgcs[color], bgcolors[C_BG1]);
	break;
      case 2:
	XSetBackground(display, cgcs[color], bgcolors[C_BG2]);
	break;
      case 3:
	XSetBackground(display, cgcs[color], bgcolors[C_BG3]);
	break;
    }
    XPutImage(display, win, cgcs[color], im,
	0, (m & 0x3f) * 8, xoff + (x << 3), yoff + (y << 3), 8, 8);
}


static void    draw_bitmap_square(ADDRESS addr)
{
    int     x, y, color;
    addr &= 0x03ff;

    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    x = addr % XSIZE;
    y = addr / XSIZE;

    color = scrbase[addr] >> 4;
    XSetBackground(display, cgcs[color], colors[scrbase[addr] & 0xf]);
    XPutImage(display, win, cgcs[color], addr & 0x200 ? im2 : im1,
	0, (addr & 0x1ff) << 3, xoff + (x << 3), yoff + (y << 3), 8, 8);
}


static void    draw_mbitmap_square(ADDRESS addr)
{
    int     x, y, color;
    addr &= 0x03ff;

    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    x = addr % XSIZE;
    y = addr / XSIZE;

    color = scrbase[addr] >> 4;
    XSetBackground(display, cgcs[color], colors[scrbase[addr] & 0xf]);
    XPutImage(display, win, cgcs[color], addr & 0x200 ? im2 : im1,
	0, (addr & 0x1ff) << 3, xoff + (x << 3), yoff + (y << 3), 8, 8);
}


/* redraws */

static void    redraw_blank(void)
{
    queuesize = 0;
    /* XClearArea(display, win, 0, 0, 0, 0, 0); */
}


static void    redraw_tilt(void)
{
    queuesize = 0;
    XFillRectangle(display, win, cgcs[0], 0, 0, XPIX, YPIX);
}


static void    redraw_text(void)
{
    BYTE   *m = scrbase;
    BYTE   *c = colorRam;
    int     x, y;

    queuesize = 0;

    im -> data = (char *) charbase;
    for (y = 0; y < YSIZE; y++)
	for (x = 0; x < XSIZE; x++) {
	    XPutImage(display, win, cgcs[*c++ & 0xf], im,
		0, *m++ * 8, xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


static void    redraw_mtext(void)
{
    int     x, y;

    queuesize = 0;
    for (y = 0; YSIZE > y; y++)
	for (x = 0; XSIZE > x; x++)
	    draw_mtext(x + y * 40);
}


static void    redraw_ectext(void)
{
    BYTE   *m = scrbase;
    BYTE   *c = colorRam;
    int     x, y, color;

    queuesize = 0;
    im -> data = (char *) charbase;
    for (y = 0; YSIZE > y; y++)
	for (x = 0; XSIZE > x; x++) {
	    color = *c++ & 0xf;
	    switch (*m >> 6) {
	      case 0:
		XSetBackground(display, cgcs[color], bgcolors[C_BG0]);
		break;
	      case 1:
		XSetBackground(display, cgcs[color], bgcolors[C_BG1]);
		break;
	      case 2:
		XSetBackground(display, cgcs[color], bgcolors[C_BG2]);
		break;
	      case 3:
		XSetBackground(display, cgcs[color], bgcolors[C_BG3]);
		break;
	    }
	    XPutImage(display, win, cgcs[color], im, 0, (*m++ & 0x3f) * 8,
		xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


static void    redraw_bitmap(void)
{
    BYTE   *m = scrbase;
    int     x, y, color, i;

    queuesize = 0;
    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    for (i = y = 0; YSIZE > y; y++)
	for (x = 0; XSIZE > x; x++, i += 8) {
	    color = *m >> 4;
	    XSetBackground(display, cgcs[color], colors[*m++ & 0xf]);
	    XPutImage(display, win, cgcs[color], i & 0x1000 ? im2 : im1,
		0, i & 0xfff, xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


static void    redraw_mbitmap(void)
{
    BYTE   *m = scrbase;
    int     x, y, color, i;

    queuesize = 0;
    im1 -> data = (char *) bitmap1_base;
    im2 -> data = (char *) bitmap2_base;
    for (i = y = 0; YSIZE > y; y++)
	for (x = 0; XSIZE > x; x++, i += 8) {
	    color = *m >> 4;
	    XSetBackground(display, cgcs[color], colors[*m++ & 0xf]);
	    XPutImage(display, win, cgcs[color], i & 0x1000 ? im2 : im1,
		0, i & 0xfff, xoff + (x << 3), yoff + (y << 3), 8, 8);
	}
}


/* bitmap draw functions */

static void    draw_chargen(ADDRESS addr)
{
    /* all chars, whose bitmap is modified must be redrawn */
    /* this does really slow the emulator down */

    int     i;
    addr = (addr >> 3) & 0xff;	/* changed char */
    for (i = 0; i < 1000; i++)
	if (scrbase[i] == addr)
	    screenqueue[queuesize++] = i;
}
